# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
import pickle
import random
from collections import defaultdict

import numpy as np
from scipy.optimize import bisect


###############################################################################
# UTILITY FUNCTIONS ###########################################################
###############################################################################
def randomhashfunction(zr, randint=random.randint):
    """Return a random hash function, mapping x in Z to ZR
    h:x -> ax + b mod R

    :param randint_function: a function that returns a random integer
    in range [a, b], including both end points. If this function is not
    given, the randint function from the standard random module is used.
    """
    bound = max(zr - 1, 1)
    a = randint(1, bound)
    b = randint(1, bound)

    def hashfunc(x):
        return (a * x + b) % zr

    return hashfunc


def count_vectorizer_func(sentences, min_n, max_n):
    """Perform a tokenization using scikit learn"""
    import sklearn
    import sklearn.feature_extraction.text as sklt

    skversion = tuple(int(x) for x in sklearn.__version__.split(".")[:2])
    if skversion < (0, 11):
        word_ngram = sklt.WordNGramAnalyzer(min_n=min_n, max_n=max)
        count_vec = sklt.CountVectorizer(analyzer=word_ngram)
    elif skversion < (0, 14):
        count_vec = sklt.CountVectorizer(min_n=min_n, max_n=max_n)
    else:
        count_vec = sklt.CountVectorizer(ngram_range=(min_n, max_n))
    # Transform and convert to lil to get rows
    data = count_vec.fit_transform(sentences).tolil()
    return [list(row) for row in data.rows], data.shape


###############################################################################
# MINHASHING ##################################################################
###############################################################################
class Minlsh:
    """Operate minhashing + locally-sensitive-hashing to find similar sentences"""

    def __init__(self, tokenizer_func=None, random_seed=None, verbose=False):
        """Initialize a minhashing/lsh object

        Parameters:
        ==========

           * tokenizer_func is a function that take the sentences
             as argument and return the rows of the sparse matrix
             of tokens, and its shape.
           * random_seed is an *optional* parameter that can be used to control
             the random seed of hashing functions. Controlling the seed allows
             to re-use the same hash functions, and which enables repeatable
             results.

           * verbose is a boolean that trigger the display of
             some informations
        """
        self._trained = False
        self._random_seed = random_seed
        self.sigmatrix = None
        if tokenizer_func:
            # Use given tokenizer_func
            self._buildmatrixdocument = lambda x, y: tokenizer_func(x)
        self._verbose = verbose

    def train(self, sentences, k=2, siglen=200):
        """Train the minlsh on the given sentences.

        - `k` is the length of the k-wordgrams used
          (the lower k is, the faster is the training)
        - `siglen` the length of the sentences signature

        """
        rows, shape = self._buildmatrixdocument(sentences, k)
        if self._verbose:
            print("Training is done. Wait while signaturing")
        self._computesignaturematrix(rows, shape, siglen)
        self._trained = True

    def _iter_wordgrams(self, sentence, k):
        """Generator of k-wordgrams on the given sentence"""
        words = sentence.split(" ")
        for r in range(len(words)):
            for width in range(1, k + 1):
                if r + width <= len(words):
                    yield " ".join(words[r : r + width])

    def _buildmatrixdocument(self, sentences, k):
        """Return a sparse matrix where :

        - Each sentence is a column
        - Each row is a element of the universal set

        Each value (r, c) is set to 1 if the element at row r is in the
        sentence c, 0 otherwise

        """
        # Use default mode
        rows, universe, sizeofuniverse = [], {}, 0
        for nb, sent in enumerate(sentences):
            row = []
            for w in self._iter_wordgrams(sent, k):
                row.append(universe.setdefault(w, sizeofuniverse))
                if row[-1] == sizeofuniverse:
                    sizeofuniverse += 1
            rows.append(row)
            if self._verbose and nb % 50000 == 0:
                print(nb)

        return rows, (len(rows), sizeofuniverse)

    def _computesignaturematrix(self, rows, shape, siglen):
        """Return a matrix where each column is the signature the document
        The signature is composed of `siglen` numbers

        The more the documents have rows in commun, the closer they are.
        """

        nrows, ncols = shape
        sig = np.empty((siglen, nrows))
        # Generate the random hash functions
        minlsh_random = random.Random(self._random_seed)
        hashfunc = [randomhashfunction(ncols, randint=minlsh_random.randint) for _ in range(siglen)]
        # Compute hashing values just for once.
        # Avoid multiple recomputations for the same column.
        hashvalues = np.array([[hashfunc[i](r) for r in range(ncols)] for i in range(siglen)])

        docind = 0
        while rows:
            doc = rows.pop(0)
            # Concatenate the needed rows.
            tmp = np.dstack([hashvalues[:, r] for r in doc])
            # Take the mininum of hashes
            sig[:, docind] = np.min(tmp[0], 1)
            docind += 1
            if self._verbose and docind % 50000 == 0:
                print((docind * 100) / nrows)
        self.sigmatrix = sig

    def save(self, savefile):
        """Save the training into `savefile` for a future use"""

        if not self._trained:
            print("Not trained, nothing to save")
            return

        with open(savefile, "wb") as fobj:
            pickler = pickle.Pickler(fobj)
            pickler.dump(self.sigmatrix)

    def load(self, savefile):
        """Load a trained minhashing"""

        with open(savefile, "rb") as fobj:
            pickler = pickle.Unpickler(fobj)
            self.sigmatrix = pickler.load()

        if self.sigmatrix is not None:
            self._trained = True
        else:
            self._trained = False

    def computebandsize(self, threshold, nbrows):
        """Compute the bandsize according to the threshold given"""

        # t ~ (1/b)^(1/r), where t is the threshold, b the number of
        # bands, and r the number of rows per band. And nbrows (the length
        # of the matrix is nbrows = b*r, so t ~ (r/L)^(1/r). So, let's
        # find the root of f(x) = (x/L)^(1/r) - t.
        def f(x):
            y = pow(x / nbrows, 1.0 / x) - threshold
            return y

        # Solve f(x) = 0, with x having values in [1, nbrows]
        return int(bisect(f, 1, nbrows))

    def predict(self, threshold, minclustersize=2):
        """Return a set of tuples of *possible* similar sentences"""
        if not self._trained:
            print("Train it before")
            return

        if not (0 < threshold <= 1):
            print("Threshold must be in ]0 ; 1]")
            return

        sig = self.sigmatrix
        # Treshold is a percent of similarity
        # It should be inverted here (0 is closed, 1 is far)
        threshold = 1 - threshold
        bandsize = self.computebandsize(threshold, self.sigmatrix.shape[0])

        buckets = defaultdict(set)
        similars = set()
        for r in range(0, sig.shape[0], bandsize):
            buckets.clear()
            for i in range(sig.shape[1]):
                buckets[tuple(sig[r : r + bandsize, i])].add(i)
            similars.update({tuple(v) for v in buckets.values() if len(v) >= minclustersize})
        return similars



""" Filters for Named Entities Recognition.
"""

from nazca.utils.dataio import sparqlquery


###############################################################################
# NER FILTERS #################################################################
###############################################################################
class AbstractNerFilter:
    """A filter used for cleaning named entities results"""

    def __call__(self, named_entities):
        raise NotImplementedError


class NerOccurenceFilter:
    """A filter based on the number of occurence of
    named entities in the results.
    """

    def __init__(self, min_occ=None, max_occ=None):
        self.min_occ = min_occ
        self.max_occ = max_occ

    def __call__(self, named_entities):
        uris = [u for u, p, t in named_entities]
        counts = {u: uris.count(u) for u in set(uris)}
        return [
            n
            for n in named_entities
            if not (
                (self.min_occ and counts[n[0]] < self.min_occ)
                or (self.max_occ and counts[n[0]] > self.max_occ)
            )
        ]


class NerRDFTypeFilter:
    """A filter based on the RDF type on entity
    E.g.

    filter = NerRDFTypeFilter('http://dbpedia.org/sparql',
                                ('http://schema.org/Place',
                                'http://dbpedia.org/ontology/Agent',
                                'http://dbpedia.org/ontology/Place'))

    """

    def __init__(self, endpoint, accepted_types):
        self.endpoint = endpoint
        self.accepted_types = accepted_types
        self.query = "SELECT ?type WHERE{<%(uri)s> rdf:type ?type}"

    def __call__(self, named_entities):
        filtered_named_entities = []
        seen_uris = {}
        for uri, p, t in named_entities:
            if uri in seen_uris:
                if seen_uris[uri]:
                    filtered_named_entities.append((uri, p, t))
            else:
                results = sparqlquery(self.endpoint, self.query % {"uri": uri})
                types = {r[0] for r in results}
                if not len(types.intersection(self.accepted_types)):
                    seen_uris[uri] = False
                else:
                    seen_uris[uri] = True
                    filtered_named_entities.append((uri, p, t))
        return filtered_named_entities


class NerDisambiguationWordParts:
    """Disambiguate named entities based on the words parts.
    E.g.:
          'toto tutu': 'http://example.com/toto_tutu',
          'toto': 'http://example.com/toto'

          Then if 'toto' is found in the text, replace the URI 'http://example.com/toto'
          by 'http://example.com/toto_tutu'
    """

    def __call__(self, named_entities):
        # Create parts dictionnary
        parts = {}
        for uri, peid, token in named_entities:
            if " " in token.word:
                for part in token.word.split(" "):
                    parts[part.lower()] = uri
        # Replace named entities
        filtered_named_entities = []
        for uri, peid, token in named_entities:
            if token.word.lower() in parts:
                # Change URI
                uri = parts[token.word.lower()]
            filtered_named_entities.append((uri, peid, token))
        return filtered_named_entities


class NerReplacementRulesFilter:
    """Allow to define replacement rules for Named Entities"""

    def __init__(self, rules):
        self.rules = rules

    def __call__(self, named_entities):
        filtered_named_entities = []
        for uri, peid, token in named_entities:
            uri = self.rules.get(uri, uri)
            filtered_named_entities.append((uri, peid, token))
        return filtered_named_entities



# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.


""" Blocking techniques.

This module implements a set of blocking techniques used to split
datasets in smaller subsets that will be aligned in more details.

Additional information:

   P. Christen, Data Matching, Data-Centric Systems and Applications,


"""
from functools import partial

from scipy.spatial import KDTree

from nazca.utils.minhashing import Minlsh
from nazca.utils.distances import soundexcode


###############################################################################
# GENERAL BLOCKING ############################################################
###############################################################################
class BaseBlocking:
    """An abstract general blocking object that exposes
    the API that should be common to all blockings object
    """

    def __init__(self, ref_attr_index, target_attr_index):
        """Build the blocking object

        Parameters
        ----------

        ref_attr_index: index of the attribute of interest in a record
                        for the reference dataset
                        (i.e. attribute to be used for key computation)

        target_attr_index: index of the attribute of interest in a record
                           for the target dataset
                           (i.e. attribute to be used for key computation)
        """
        self.ref_attr_index = ref_attr_index
        self.target_attr_index = target_attr_index
        self.refids = None
        self.targetids = None
        self.is_fitted = False

    def _fit(self, refset, targetset):
        raise NotImplementedError

    def _iter_blocks(self):
        """Internal iteration function over blocks"""
        raise NotImplementedError

    def _cleanup(self):
        """Internal cleanup blocking for further use (e.g. in pipeline)"""
        raise NotImplementedError

    def fit(self, refset, targetset):
        """Fit the blocking technique on the reference and target datasets

        Parameters
        ----------
        refset: a dataset (list of records)

        targetset: a dataset (list of records)
        """
        self._fit(refset, targetset)
        # Keep ids for blocks building
        self.refids = [(i, r[0]) for i, r in enumerate(refset)]
        self.targetids = [(i, r[0]) for i, r in enumerate(targetset)]
        self.is_fitted = True

    def iter_blocks(self):
        """Iterator over the different possible blocks.

        Returns
        -------

        (block1, block2): The blocks are always (reference_block, target_block)
                          and contains the pair (index, id) of the record in the
                          corresponding dataset.
        """
        assert self.is_fitted
        return self._iter_blocks()

    def iter_indice_blocks(self):
        """Iterator over the different possible blocks.

        Returns
        -------

        (block1, block2): The blocks are always (reference_block, target_block)
                          and contains the indexes of the record in the
                          corresponding dataset.
        """
        assert self.is_fitted
        for block1, block2 in self._iter_blocks():
            yield [r[0] for r in block1], [r[0] for r in block2]

    def iter_id_blocks(self):
        """Iterator over the different possible blocks.

        Returns
        -------

        (block1, block2): The blocks are always (reference_block, target_block)
                          and contains the ids of the record in the
                          corresponding dataset.
        """
        assert self.is_fitted
        for block1, block2 in self._iter_blocks():
            yield [r[1] for r in block1], [r[1] for r in block2]

    def iter_pairs(self):
        """Iterator over the different possible pairs.

        Returns
        -------

        (pair1, pari2): The pairs are always ((ind_reference, id_reference),
                                              (ind_target, id_target))
                        and are the ids of the record in the corresponding dataset.
        """
        assert self.is_fitted
        for block1, block2 in self.iter_blocks():
            for val1 in block1:
                for val2 in block2:
                    yield val1, val2

    def iter_indice_pairs(self):
        """Iterator over the different possible pairs.

        Returns
        -------

        (pair1, pari2): The pairs are always (ind_reference, ind_target)
                        and are the ids of the record in the corresponding dataset.
        """
        assert self.is_fitted
        for block1, block2 in self.iter_indice_blocks():
            for val1 in block1:
                for val2 in block2:
                    yield val1, val2

    def iter_id_pairs(self):
        """Iterator over the different possible pairs.

        Returns
        -------

        (pair1, pari2): The pairs are always (id_reference, id_target)
                        and are the ids of the record in the corresponding dataset.
        """
        assert self.is_fitted
        for block1, block2 in self.iter_id_blocks():
            for val1 in block1:
                for val2 in block2:
                    yield val1, val2

    def cleanup(self):
        """Cleanup blocking for further use (e.g. in pipeline)"""
        self.is_fitted = True
        self._cleanup()


###############################################################################
# KEY BLOCKING ################################################################
###############################################################################
class KeyBlocking(BaseBlocking):
    """This blocking technique is based on a a blocking criteria
    (or blocking key), that will be used to divide the datasets.

    The main idea here is:

    1 - to create an index of f(x) for each x in the reference set.

    2 - to create an index of f(y) for each y in the target set.

    3 - to iterate on each distinct value of f(x) and to return
        the identifiers of the records of the both sets for this value.
    """

    def __init__(self, ref_attr_index, target_attr_index, callback, ignore_none=False):
        super().__init__(ref_attr_index, target_attr_index)
        self.callback = callback
        self.ignore_none = ignore_none
        self.reference_index = {}
        self.target_index = {}

    def _fit(self, refset, targetset):
        """Fit a dataset in an index using the callback"""
        for ind, rec in enumerate(refset):
            key = self.callback(rec[self.ref_attr_index])
            if not key and self.ignore_none:
                continue
            self.reference_index.setdefault(key, []).append((ind, rec[0]))
        for ind, rec in enumerate(targetset):
            key = self.callback(rec[self.target_attr_index])
            if not key and self.ignore_none:
                continue
            self.target_index.setdefault(key, []).append((ind, rec[0]))

    def _iter_blocks(self):
        """Iterator over the different possible blocks.

        Returns
        -------

        (block1, block2): The blocks are always (reference_block, target_block)
                          and containts the indexes of the record in the
                          corresponding dataset.
        """
        for key, block1 in self.reference_index.items():
            block2 = self.target_index.get(key)
            if block1 and block2:
                yield (block1, block2)

    def _cleanup(self):
        """Cleanup blocking for further use (e.g. in pipeline)"""
        self.reference_index = {}
        self.target_index = {}


class SoundexBlocking(KeyBlocking):
    def __init__(
        self,
        ref_attr_index,
        target_attr_index,
        language="french",
    ):
        super().__init__(ref_attr_index, target_attr_index, partial(soundexcode, language=language))


###############################################################################
# BIGRAM BLOCKING #############################################################
###############################################################################
class NGramBlocking(BaseBlocking):
    """This blocking technique is based on a a n-gram key."""

    def __init__(self, ref_attr_index, target_attr_index, ngram_size=2, depth=2):
        super().__init__(ref_attr_index, target_attr_index)
        self.ngram_size = ngram_size
        self.depth = depth
        self.reference_index = {}
        self.target_index = {}

    def _fit_dataset(self, dataset, cur_index, attr_index):
        """Fit a dataset"""
        for ind, r in enumerate(dataset):
            cur_dict = cur_index
            text = r[attr_index]
            for i in range(self.depth):
                ngram = text[i * self.ngram_size : (i + 1) * self.ngram_size]
                if i < self.depth - 1:
                    cur_dict = cur_dict.setdefault(ngram, {})
            cur_dict.setdefault(ngram, []).append((ind, r[0]))

    def _fit(self, refset, targetset):
        """Fit the two sets (reference set and target set)"""
        self._fit_dataset(refset, self.reference_index, self.ref_attr_index)
        self._fit_dataset(targetset, self.target_index, self.target_attr_index)

    def _iter_dict(self, ref_cur_dict, target_cur_dict):
        """Iterative function used to create blocks from dicts"""
        for key, sub_dict in ref_cur_dict.items():
            if key in target_cur_dict:
                if isinstance(sub_dict, dict):
                    # There is another dict layer
                    yield from self._iter_dict(sub_dict, target_cur_dict[key])
                else:
                    # This is a list
                    yield sub_dict, target_cur_dict[key]

    def _iter_blocks(self):
        """Iterator over the different possible blocks.

        Returns
        -------

        (block1, block2): The blocks are always (reference_block, target_block)
                          and containts the indexes of the record in the
                          corresponding dataset.
        """
        for block1, block2 in self._iter_dict(self.reference_index, self.target_index):
            if block1 and block2:
                yield block1, block2

    def _cleanup(self):
        """Cleanup blocking for further use (e.g. in pipeline)"""
        self.reference_index = {}
        self.target_index = {}


###############################################################################
# SORTKEY BLOCKING ############################################################
###############################################################################
class SortedNeighborhoodBlocking(BaseBlocking):
    """This blocking technique is based on a a sorting blocking criteria
    (or blocking key), that will be used to divide the datasets.
    """

    def __init__(self, ref_attr_index, target_attr_index, key_func=lambda x: x, window_width=20):
        super().__init__(ref_attr_index, target_attr_index)
        self.key_func = key_func
        self.window_width = window_width
        self.sorted_dataset = None

    def _fit(self, refset, targetset):
        """Fit a dataset in an index using the callback"""
        self.sorted_dataset = [
            ((ind, r[0]), r[self.ref_attr_index], 0) for ind, r in enumerate(refset)
        ]
        self.sorted_dataset.extend(
            [((ind, r[0]), r[self.target_attr_index], 1) for ind, r in enumerate(targetset)]
        )
        self.sorted_dataset.sort(key=lambda x: self.key_func(x[1]))

    def _iter_blocks(self):
        """Iterator over the different possible blocks."""
        for ind, (rid, record, dset) in enumerate(self.sorted_dataset):
            # Only keep reference set record
            if dset == 1:
                continue
            block1 = [
                rid,
            ]
            minind = ind - self.window_width
            minind = minind if minind >= 0 else 0
            maxind = ind + self.window_width + 1
            block2 = [ri for ri, re, d in self.sorted_dataset[minind:maxind] if d == 1]
            if block1 and block2:
                yield (block1, block2)

    def _cleanup(self):
        """Cleanup blocking for further use (e.g. in pipeline)"""
        self.sorted_dataset = None


###############################################################################
# MERGE BLOCKING ##############################################################
###############################################################################
class MergeBlocking(BaseBlocking):
    """This blocking technique keep only one appearance of one given values,
    and removes all the other records having this value.
    The merge is based on a score function

    E.g.
      ('http://fr.wikipedia.org/wiki/Paris_%28Texas%29', 'Paris', 25898)
      ('http://fr.wikipedia.org/wiki/Paris', 'Paris', 12223100)

    could be (with a score function based on the population (third value):

      ('http://fr.wikipedia.org/wiki/Paris', 'Paris', 12223100)

    !!! WARNING !!! This is only done on ONE set (the one with a non null attr index)
    """

    def __init__(self, ref_attr_index, target_attr_index, score_func):
        super().__init__(ref_attr_index, target_attr_index)
        self.score_func = score_func
        self.merged_dataset = None
        self.other_dataset = None
        if ref_attr_index is None and target_attr_index is None:
            raise ValueError(
                "At least one of ref_attr_index or target_attr_index " "should not be None"
            )

    def _fit(self, refset, targetset):
        """Fit a dataset in an index using the callback"""
        if self.ref_attr_index is not None:
            # Merge refset
            self.merged_dataset = self._merge_dataset(refset, self.ref_attr_index)
            self.other_dataset = [(ind, r[0]) for ind, r in enumerate(targetset)]
        else:
            # Merge targetset
            self.merged_dataset = self._merge_dataset(targetset, self.target_attr_index)
            self.other_dataset = [(ind, r[0]) for ind, r in enumerate(refset)]

    def _merge_dataset(self, dataset, attr_index):
        """Merge a dataset"""
        merged_dataset_dict = {}
        for ind, record in enumerate(dataset):
            score = self.score_func(record)
            if record[attr_index] not in merged_dataset_dict:
                # Create new entry
                merged_dataset_dict[record[attr_index]] = (ind, record, score)
            elif (
                record[attr_index] in merged_dataset_dict
                and merged_dataset_dict[record[attr_index]][2] < score
            ):
                # Change current score
                merged_dataset_dict[record[attr_index]] = (ind, record, score)
        return [(ind, r[0]) for ind, r, score in merged_dataset_dict.values()]

    def _iter_blocks(self):
        """Iterator over the different possible blocks."""
        if self.ref_attr_index is not None:
            yield self.merged_dataset, self.other_dataset
        else:
            # self.target_attr_index is not None
            yield self.other_dataset, self.merged_dataset

    def _cleanup(self):
        """Cleanup blocking for further use (e.g. in pipeline)"""
        self.merged_dataset = None
        self.other_dataset = None


###############################################################################
# CLUSTERING-BASED BLOCKINGS ##################################################
###############################################################################
class KmeansBlocking(BaseBlocking):
    """A blocking technique based on Kmeans"""

    def __init__(self, ref_attr_index, target_attr_index, n_clusters=None):
        super().__init__(ref_attr_index, target_attr_index)
        self.n_clusters = n_clusters
        self.kmeans = None
        self.predicted = None
        from sklearn import cluster

        self.cluster_class = cluster.KMeans

    def _fit(self, refset, targetset):
        """Fit the reference dataset."""
        # If an element is None (missing), use instead the identity element.
        # The identity element is defined as the 0-vector
        idelement = tuple(
            [
                0,
            ]
            * len(refset[0][self.ref_attr_index])
        )
        # We assume here that there are at least 2 elements in the refset
        n_clusters = self.n_clusters or (len(refset) // 10 or len(refset) // 2)
        try:
            kmeans = self.cluster_class(n_clusters=n_clusters)
        except TypeError:
            # Try older API version of sklearn
            kmeans = self.cluster_class(k=n_clusters)
        kmeans.fit([elt[self.ref_attr_index] or idelement for elt in refset])
        self.kmeans = kmeans
        # Predict on targetset
        self.predicted = self.kmeans.predict(
            [elt[self.target_attr_index] or idelement for elt in targetset]
        )

    def _iter_blocks(self):
        """Iterator over the different possible blocks.

        Returns
        -------

        (block1, block2): The blocks are always (reference_block, target_block)
                          and containts the indexes of the record in the
                          corresponding dataset.
        """
        n_clusters = self.kmeans.n_clusters if hasattr(self.kmeans, "n_clusters") else self.kmeans.k
        neighbours = [[[], []] for _ in range(n_clusters)]
        for ind, li in enumerate(self.predicted):
            neighbours[li][1].append(self.targetids[ind])
        for ind, li in enumerate(self.kmeans.labels_):
            neighbours[li][0].append(self.refids[ind])
        for block1, block2 in neighbours:
            if len(block1) and len(block2):
                yield block1, block2

    def _cleanup(self):
        """Cleanup blocking for further use (e.g. in pipeline)"""
        self.kmeans = None
        self.predicted = None


###############################################################################
# KDTREE BLOCKINGS ############################################################
###############################################################################
class KdTreeBlocking(BaseBlocking):
    """A blocking technique based on KdTree"""

    def __init__(self, ref_attr_index, target_attr_index, threshold=0.1):
        super().__init__(ref_attr_index, target_attr_index)
        self.threshold = threshold
        self.reftree = None
        self.targettree = None
        self.nb_elements = None

    def _fit(self, refset, targetset):
        """Fit the blocking"""
        firstelement = refset[0][self.ref_attr_index]
        self.nb_elements = len(refset)
        idsize = len(firstelement) if isinstance(firstelement, (tuple, list)) else 1
        idelement = (0,) * idsize
        # KDTree is expecting a two-dimensional array
        if idsize == 1:
            self.reftree = KDTree([(elt[self.ref_attr_index],) or idelement for elt in refset])
            self.targettree = KDTree(
                [(elt[self.target_attr_index],) or idelement for elt in targetset]
            )
        else:
            self.reftree = KDTree([elt[self.ref_attr_index] or idelement for elt in refset])
            self.targettree = KDTree(
                [elt[self.target_attr_index] or idelement for elt in targetset]
            )

    def _iter_blocks(self):
        """Iterator over the different possible blocks.

        Returns
        -------

        (block1, block2): The blocks are always (reference_block, target_block)
                          and containts the indexes of the record in the
                          corresponding dataset.
        """
        extraneighbours = self.reftree.query_ball_tree(self.targettree, self.threshold)
        neighbours = []
        for ind in range(self.nb_elements):
            if not extraneighbours[ind]:
                continue
            _ref = [
                self.refids[ind],
            ]
            _target = [self.targetids[v] for v in extraneighbours[ind]]
            neighbours.append((_ref, _target))
        for block1, block2 in neighbours:
            if len(block1) and len(block2):
                yield block1, block2

    def _cleanup(self):
        """Cleanup blocking for further use (e.g. in pipeline)"""
        self.reftree = None
        self.targettree = None
        self.nb_elements = None


###############################################################################
# MINHASHING BLOCKINGS ########################################################
###############################################################################
class MinHashingBlocking(BaseBlocking):
    """A blocking technique based on MinHashing"""

    # This is the random seed to be used to computed the hash functions, when
    # the user wants repeatable results.
    FIXED_RANDOM_SEED = "4; chosen by fair dice roll. Guaranteed to be random."

    def __init__(
        self,
        ref_attr_index,
        target_attr_index,
        threshold=0.1,
        kwordsgram=1,
        siglen=200,
        repeatable=False,
    ):
        """Specific parameters of the MinHashingBlocking:

        :param kwordsgram: is the length of the k-wordgrams used
            (the lower k is, the faster is the training)
        :param siglen: the length of the sentences signature
        :param repeatable: whether a fixed random seed should be used or not.
            Using a fixed random seed allows to have repeatable results.
        """

        super().__init__(ref_attr_index, target_attr_index)
        self.threshold = threshold
        self.kwordsgram = kwordsgram
        self.siglen = siglen
        self._repeatable = repeatable
        self.minhasher = self._get_new_minhasher()
        self.nb_elements = None

    def _get_new_minhasher(self):
        if self._repeatable:
            return Minlsh(random_seed=self.FIXED_RANDOM_SEED)
        return Minlsh()

    def _fit(self, refset, targetset):
        """Find the blocking using minhashing"""
        # If an element is None (missing), use instead the identity element.
        idelement = ""
        self.minhasher.train(
            [elt[self.ref_attr_index] or idelement for elt in refset]
            + [elt[self.target_attr_index] or idelement for elt in targetset],
            self.kwordsgram,
            self.siglen,
        )
        self.nb_elements = len(refset)

    def _iter_blocks(self):
        """Iterator over the different possible blocks.

        Returns
        -------

        (block1, block2): The blocks are always (reference_block, target_block)
                          and containts the indexes of the record in the
                          corresponding dataset.
        """
        rawneighbours = self.minhasher.predict(self.threshold)
        neighbours = []
        for data in rawneighbours:
            neighbours.append([[], []])
            for i in data:
                if i >= self.nb_elements:
                    neighbours[-1][1].append(self.targetids[i - self.nb_elements])
                else:
                    neighbours[-1][0].append(self.refids[i])
            if len(neighbours[-1][0]) == 0 or len(neighbours[-1][1]) == 0:
                neighbours.pop()
        for block1, block2 in neighbours:
            if len(block1) and len(block2):
                yield block1, block2

    def _cleanup(self):
        """Cleanup blocking for further use (e.g. in pipeline)"""
        self.minhasher = self._get_new_minhasher()
        self.nb_elements = None


###############################################################################
# BLOCKING PIPELINE ###########################################################
###############################################################################
class PipelineBlocking(BaseBlocking):
    """Pipeline multiple blocking techniques"""

    def __init__(self, blockings, collect_stats=False):
        """Build the blocking object

        Parameters
        ----------

        blockings: ordered list of blocking objects
        """
        self.blockings = blockings
        self.stored_blocks = []
        self.collect_stats = collect_stats
        self.stats = {}

    def _fit(self, refset, targetset):
        """Internal fit of the pipeline"""
        self._recursive_fit(refset, targetset, range(len(refset)), range(len(targetset)), 0)

    def _recursive_fit(self, refset, targetset, ref_index, target_index, ind):
        """Recursive fit of the blockings.
        Blocks are stored in the stored_blocks attribute.
        """
        if ind < len(self.blockings) - 1:
            # There are other blockings after this one
            blocking = self.blockings[ind]
            blocking.cleanup()
            blocking.fit([refset[i] for i in ref_index], [targetset[i] for i in target_index])
            for block1, block2 in blocking.iter_indice_blocks():
                ind_block1 = [ref_index[i] for i in block1]
                ind_block2 = [target_index[i] for i in block2]
                if self.collect_stats:
                    self.stats.setdefault(ind, []).append((len(block1), len(block2)))
                self._recursive_fit(refset, targetset, ind_block1, ind_block2, ind + 1)
        else:
            # This is the final blocking
            blocking = self.blockings[ind]
            blocking.cleanup()
            blocking.fit([refset[i] for i in ref_index], [targetset[i] for i in target_index])
            for block1, block2 in blocking.iter_blocks():
                ind_block1 = [(ref_index[i], _id) for i, _id in block1]
                ind_block2 = [(target_index[i], _id) for i, _id in block2]
                if self.collect_stats:
                    self.stats.setdefault(ind, []).append((len(block1), len(block2)))
                self.stored_blocks.append((ind_block1, ind_block2))

    def _iter_blocks(self):
        """Internal iteration function over blocks"""
        for block1, block2 in self.stored_blocks:
            if block1 and block2:
                yield block1, block2



#
# copyright 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

from nazca.ner import NerProcess
from nazca.ner.filters import (
    NerOccurenceFilter,
    NerReplacementRulesFilter,
    NerDisambiguationWordParts,
)
from nazca.ner.sources import NerSourceLexicon
from nazca.utils.tokenizer import Token, Sentence, NLTK_AVAILABLE


@unittest.skipUnless(NLTK_AVAILABLE, "nltk is not available")
class FilterTest(unittest.TestCase):
    """Test of filters"""

    def test_occurence_filter_min_occ(self):
        """Test occurence filter"""
        text = "Hello everyone, this is   me speaking. And me."
        source1 = NerSourceLexicon(
            {"everyone": "http://example.com/everyone", "me": "http://example.com/me"}
        )
        source2 = NerSourceLexicon({"me": "http://example2.com/me"})
        _filter = NerOccurenceFilter(min_occ=2)
        ner = NerProcess((source1, source2), filters=(_filter,))
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=26, end=28, sentence=Sentence(indice=0, start=0, end=38)
                    ),
                ),
                (
                    "http://example2.com/me",
                    None,
                    Token(
                        word="me", start=26, end=28, sentence=Sentence(indice=0, start=0, end=38)
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=43, end=45, sentence=Sentence(indice=1, start=39, end=46)
                    ),
                ),
                (
                    "http://example2.com/me",
                    None,
                    Token(
                        word="me", start=43, end=45, sentence=Sentence(indice=1, start=39, end=46)
                    ),
                ),
            ],
        )

    def test_occurence_filter_max_occ(self):
        """Test occurence filter"""
        text = "Hello everyone, this is   me speaking. And me."
        source1 = NerSourceLexicon(
            {"everyone": "http://example.com/everyone", "me": "http://example.com/me"}
        )
        source2 = NerSourceLexicon({"me": "http://example2.com/me"})
        _filter = NerOccurenceFilter(max_occ=1)
        ner = NerProcess((source1, source2), filters=(_filter,))
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/everyone",
                    None,
                    Token(
                        word="everyone",
                        start=6,
                        end=14,
                        sentence=Sentence(indice=0, start=0, end=38),
                    ),
                ),
            ],
        )

    def test_disambiguation_word_length(self):
        """Test occurence filter"""
        text = "Hello toto tutu. And toto."
        source = NerSourceLexicon(
            {"toto tutu": "http://example.com/toto_tutu", "toto": "http://example.com/toto"}
        )
        _filter = NerDisambiguationWordParts()
        ner = NerProcess((source,), filters=(_filter,))
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/toto_tutu",
                    None,
                    Token(
                        word="toto tutu",
                        start=6,
                        end=15,
                        sentence=Sentence(indice=0, start=0, end=16),
                    ),
                ),
                (
                    "http://example.com/toto_tutu",
                    None,
                    Token(
                        word="toto", start=21, end=25, sentence=Sentence(indice=1, start=17, end=26)
                    ),
                ),
            ],
        )

    def test_disambiguation_word_case(self):
        """Test occurence filter"""
        text = "Hello Toto Tutu. And Toto."
        source = NerSourceLexicon(
            {"Toto Tutu": "http://example.com/toto_tutu", "Toto": "http://example.com/toto"}
        )
        _filter = NerDisambiguationWordParts()
        ner = NerProcess((source,), filters=(_filter,))
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/toto_tutu",
                    None,
                    Token(
                        word="Toto Tutu",
                        start=6,
                        end=15,
                        sentence=Sentence(indice=0, start=0, end=16),
                    ),
                ),
                (
                    "http://example.com/toto_tutu",
                    None,
                    Token(
                        word="Toto", start=21, end=25, sentence=Sentence(indice=1, start=17, end=26)
                    ),
                ),
            ],
        )

    def test_rules_filter(self):
        """Test rules filter"""
        text = "Hello toto tutu. And toto."
        source = NerSourceLexicon(
            {"toto tutu": "http://example.com/toto_tutu", "toto": "http://example.com/toto"}
        )
        rules = {"http://example.com/toto": "http://example.com/tata"}
        _filter = NerReplacementRulesFilter(rules)
        ner = NerProcess((source,), filters=(_filter,))
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/toto_tutu",
                    None,
                    Token(
                        word="toto tutu",
                        start=6,
                        end=15,
                        sentence=Sentence(indice=0, start=0, end=16),
                    ),
                ),
                (
                    "http://example.com/tata",
                    None,
                    Token(
                        word="toto", start=21, end=25, sentence=Sentence(indice=1, start=17, end=26)
                    ),
                ),
            ],
        )


if __name__ == "__main__":
    unittest.main()



""" Data for Nazca
"""

import os.path as osp

HERE = osp.join(osp.abspath(osp.dirname(__file__)))
with open(osp.join(HERE, "french_lemmas.txt")) as fobj:
    FRENCH_LEMMAS = dict(t.strip().split("\t") for t in fobj)



#
# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
from functools import partial
from os import path

from nazca.utils.normalize import simplify
from nazca.utils.minhashing import Minlsh, count_vectorizer_func
from nazca.data import FRENCH_LEMMAS

TESTDIR = path.dirname(__file__)


class MinLSHTest(unittest.TestCase):
    def test_iter_wordgrams(self):
        sentence = "nom de la rose"
        minlsh = Minlsh()
        results = list(minlsh._iter_wordgrams(sentence, 2))
        truth = ["nom de", "nom", "de la", "de", "la rose", "la", "rose"]
        self.assertEqual(len(results), len(truth))
        self.assertEqual(set(results), set(truth))

    def test_iter_wordgrams_sklearn(self):
        sentences = ("nom de la rose", "nom de la")
        tokenizer_func = partial(count_vectorizer_func, min_n=1, max_n=2)
        minlsh = Minlsh(tokenizer_func=tokenizer_func)
        rows, shape = list(minlsh._buildmatrixdocument(sentences, 2))
        self.assertEqual(shape, (2, 7))
        self.assertEqual(rows[0], [0, 1, 2, 3, 4, 5, 6])
        self.assertEqual(rows[1], [0, 1, 2, 4, 5])

    def test_all(self):
        sentences = [
            "Un nuage flotta dans le grand ciel bleu.",
            "Des grands nuages noirs flottent dans le ciel.",
            "Je n'aime pas ce genre de bandes dessinées tristes.",
            "J'aime les bandes dessinées de genre comiques.",
            "Pour quelle occasion vous êtes-vous apprêtée ?",
            "Je les vis ensemble à plusieurs occasions.",
            "Je les ai vus ensemble à plusieurs occasions.",
        ]
        minlsh = Minlsh(random_seed="spam eggs bacon")
        minlsh.train(
            (simplify(s, FRENCH_LEMMAS, remove_stopwords=True) for s in sentences),
            1,
            200,
        )
        self.assertEqual({(0, 1), (2, 3), (5, 6)}, minlsh.predict(0.4))


if __name__ == "__main__":
    unittest.main()



#
# copyright 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

from nazca.utils.tokenizer import RichStringTokenizer, Token, Sentence, NLTK_AVAILABLE


@unittest.skipUnless(NLTK_AVAILABLE, "nltk is not available")
class TokenizerTest(unittest.TestCase):
    """Test of tokenizer"""

    def test_find_sentences(self):
        text = "Hello everyone, this is   me speaking. And me."
        sentences = RichStringTokenizer.find_sentences(text)
        self.assertEqual(sentences[0], Sentence(indice=0, start=0, end=38))
        self.assertEqual(sentences[1], Sentence(indice=1, start=39, end=46))

    def test_richstringtokenizer(self):
        text = "Hello everyone, this is   me speaking. And me."
        tokenizer = RichStringTokenizer(text, token_min_size=1, token_max_size=3)
        tokens = list(tokenizer)
        self.assertEqual(len(tokens), 18)
        t1 = Token(
            word="Hello everyone this",
            start=0,
            end=20,
            sentence=Sentence(indice=0, start=0, end=38),
        )
        self.assertEqual(tokens[0], t1)
        t2 = Token(word="And", start=39, end=42, sentence=Sentence(indice=1, start=39, end=46))
        self.assertEqual(tokens[16], t2)

    def test_richstringtokenizer_loadtext(self):
        text = "Hello everyone, this is   me speaking. And me."
        tokenizer = RichStringTokenizer(text, token_min_size=1, token_max_size=3)
        tokens = list(tokenizer)
        self.assertEqual(len(tokens), 18)
        tokenizer.load_text("Hello everyone")
        tokens = list(tokenizer)
        self.assertEqual(len(tokens), 3)

    def test_richstringtokenizer_minsize(self):
        text = "Hello everyone, this is   me speaking. And me."
        tokenizer = RichStringTokenizer(text, token_min_size=2, token_max_size=3)
        tokens = list(tokenizer)
        self.assertEqual(len(tokens), 10)
        t1 = Token(
            word="me speaking", start=26, end=37, sentence=Sentence(indice=0, start=0, end=38)
        )
        self.assertEqual(tokens[8], t1)

    def test_richstringtokenizer_maxsize(self):
        text = "Hello everyone, this is   me speaking. And me."
        tokenizer = RichStringTokenizer(text, token_min_size=1, token_max_size=4)
        tokens = list(tokenizer)
        self.assertEqual(len(tokens), 21)
        t1 = Token(word="And me", start=39, end=45, sentence=Sentence(indice=1, start=39, end=46))
        self.assertEqual(tokens[18], t1)

    def test_richstringtokenizer_sentences(self):
        text = "Hello everyone, this is   me speaking. And me ! Why not me ? Blup"
        tokenizer = RichStringTokenizer(text, token_min_size=1, token_max_size=4)
        sentences = tokenizer.find_sentences(text)
        self.assertEqual(len(sentences), 4)
        self.assertEqual(
            text[sentences[0].start : sentences[0].end], "Hello everyone, this is   me speaking."
        )
        self.assertEqual(text[sentences[1].start : sentences[1].end], "And me !")
        self.assertEqual(text[sentences[2].start : sentences[2].end], "Why not me ?")
        self.assertEqual(text[sentences[3].start : sentences[3].end], "Blup")


if __name__ == "__main__":
    unittest.main()



# See http://en.wikipedia.org/wiki/List_of_U.S._state_abbreviations
# WARNING: The name of each state should be in French
# (e.g. "Floride", not "Florida")
US_STATES = {
    "AK": "Alaska",
    "AL": "Alabama",
    "AR": "Arkansas",
    "AZ": "Arizona",
    "Ala.": "Alabama",
    "Alas.": "Alaska",
    "Alaska": "Alaska",
    "Ariz.": "Arizona",
    "Ark.": "Arkansas",
    "Az.": "Arizona",
    "CA": "Californie",
    "CF": "Californie",
    "CL": "Colorado",
    "CO": "Colorado",
    "CT": "Connecticut",
    "Ca.": "Californie",
    "Cal.": "Californie",
    "Cali.": "Californie",
    "Calif.": "Californie",
    "Col.": "Colorado",
    "Colo.": "Colorado",
    "Conn.": "Connecticut",
    "Ct.": "Connecticut",
    "D.C.": "District of ColuFederal district",
    "DC": "District of ColuFederal district",
    "DE": "Delaware",
    "DL": "Delaware",
    "De.": "Delaware",
    "Del.": "Delaware",
    "FL": "Floride",
    "Fl.": "Floride",
    "Fla.": "Floride",
    "Flor.": "Floride",
    "GA": "Géorgie",
    "Ga.": "Géorgie",
    "H.I.": "Hawaii",
    "HA": "Hawaii",
    "HI": "Hawaii",
    "Hawaii": "Hawaii",
    "IA": "Iowa",
    "ID": "Idaho",
    "IL": "Illinois",
    "IN": "Indiana",
    "Ia.": "Iowa",
    "Id.": "Idaho",
    "Ida.": "Idaho",
    "Idaho": "Idaho",
    "Il.": "Illinois",
    "Ill's": "Illinois",
    "Ill.": "Illinois",
    "Ills.": "Illinois",
    "In.": "Indiana",
    "Ind.": "Indiana",
    "Ioa.": "Iowa",
    "Iowa": "Iowa",
    "KA": "Kansas",
    "KS": "Kansas",
    "KY": "Kentucky",
    "Ka.": "Kansas",
    "Kan.": "Kansas",
    "Kans.": "Kansas",
    "Ks.": "Kansas",
    "Ky.": "Kentucky",
    "LA": "Louisiane",
    "La.": "Louisiane",
    "MA": "Massachusetts",
    "MC": "Michigan",
    "MD": "Maryland",
    "ME": "Maine",
    "MI": "Mississippi",
    "MN": "Minnesota",
    "MO": "Missouri",
    "MS": "Mississippi",
    "MT": "Montana",
    "Maine": "Maine",
    "Mass.": "Massachusetts",
    "Md.": "Maryland",
    "Me.": "Maine",
    "Mich.": "Michigan",
    "Minn.": "Minnesota",
    "Miss.": "Mississippi",
    "Mn.": "Minnesota",
    "Mo.": "Missouri",
    "Mont.": "Montana",
    "N. Car.": "Caroline du Nord",
    "N. Dak.": "Dakota du Nord",
    "N. Mex.": "Nouveau-Mexique",
    "N. York": "New York",
    "N.C.": "Caroline du Nord",
    "N.D.": "Dakota du Nord",
    "N.H.": "New Hampshire",
    "N.J.": "New Jersey",
    "N.M.": "Nouveau-Mexique",
    "N.Y.": "New York",
    "NB": "Nebraska",
    "NC": "Caroline du Nord",
    "ND": "Dakota du Nord",
    "NE": "Nebraska",
    "NH": "New Hampshire",
    "NJ": "New Jersey",
    "NM": "Nouveau-Mexique",
    "NV": "Nevada",
    "NY": "New York",
    "Neb.": "Nebraska",
    "Nebr.": "Nebraska",
    "Nev.": "Nevada",
    "New M.": "Nouveau-Mexique",
    "NoDak": "Dakota du Nord",
    "Nv.": "Nevada",
    "O.": "Ohio",
    "OH": "Ohio",
    "OK": "Oklahoma",
    "OR": "Oregon",
    "Oh.": "Ohio",
    "Ohio": "Ohio",
    "Ok.": "Oklahoma",
    "Okla.": "Oklahoma",
    "Or.": "Oregon",
    "Ore.": "Oregon",
    "Oreg.": "Oregon",
    "PA": "Pennsylvanie",
    "Pa.": "Pennsylvanie",
    "R.I.": "Rhode Island",
    "R.I. & P.P.": "Rhode Island",
    "RI": "Rhode Island",
    "S. Car.": "Caroline du Sud",
    "S. Dak.": "Dakota du Sud",
    "S.C.": "Caroline du Sud",
    "S.D.": "Dakota du Sud",
    "SC": "Caroline du Sud",
    "SD": "Dakota du Sud",
    "SoDak": "Dakota du Sud",
    "State": "Utah",
    "TN": "Tennessee",
    "TX": "Texas",
    "Tenn.": "Tennessee",
    "Tex.": "Texas",
    "Texas": "Texas",
    "Tn.": "Tennessee",
    "Tx.": "Texas",
    "US-AL": "Alabama",
    "US-AR": "Arkansas",
    "US-AZ": "Arizona",
    "US-CA": "Californie",
    "US-CO": "Colorado",
    "US-CT": "Connecticut",
    "US-DC": "District of ColuFederal district",
    "US-DE": "Delaware",
    "US-FL": "Floride",
    "US-GA": "Géorgie",
    "US-IL": "Illinois",
    "US-IN": "Indiana",
    "US-KY": "Kentucky",
    "US-LA": "Louisiane",
    "US-MA": "Massachusetts",
    "US-MD": "Maryland",
    "US-MI": "Michigan",
    "US-MN": "Minnesota",
    "US-MO": "Missouri",
    "US-MS": "Mississippi",
    "US-MT": "Montana",
    "US-NC": "Caroline du Nord",
    "US-ND": "Dakota du Nord",
    "US-NE": "Nebraska",
    "US-NH": "New Hampshire",
    "US-NJ": "New Jersey",
    "US-NM": "Nouveau-Mexique",
    "US-NY": "New York",
    "US-OK": "Oklahoma",
    "US-PA": "Pennsylvanie",
    "US-RI": "Rhode Island",
    "US-SC": "Caroline du Sud",
    "US-SD": "Dakota du Sud",
    "US-TN": "Tennessee",
    "US-VA": "Virginia",
    "US-VT": "Vermont",
    "US-WA": "Washington",
    "US-WI": "Wisconsin",
    "US-WV": "Virginie occidentale",
    "US-WY": "Wyoming",
    "UT": "Utah",
    "Ut.": "Utah",
    "Utah": "Utah",
    "VA": "Virginia",
    "VT": "Vermont",
    "Va.": "Virginia",
    "Vt.": "Vermont",
    "W. Va.": "Virginie occidentale",
    "W. Virg.": "Virginie occidentale",
    "W.V.": "Virginie occidentale",
    "W.Va.": "Virginie occidentale",
    "WA": "Washington",
    "WI": "Wisconsin",
    "WN": "Washington",
    "WS": "Wisconsin",
    "WV": "Virginie occidentale",
    "WY": "Wyoming",
    "Wa.": "Washington",
    "Wash.": "Washington",
    "Wash. D.C.": "District of ColuFederal district",
    "Wi.": "Wisconsin",
    "Wis.": "Wisconsin",
    "Wisc.": "Wisconsin",
    "Wn.": "Washington",
    "Wy.": "Wyoming",
    "Wyo.": "Wyoming",
}



#
# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest
import random

from os import path

import numpy as np

import nazca.rl.aligner as alig
import nazca.rl.blocking as blo
from nazca.utils.distances import LevenshteinProcessing, GeographicalProcessing

# Make sure tests are repeatable
random.seed(6)


TESTDIR = path.dirname(__file__)


class AlignerTestCase(unittest.TestCase):
    def test_align(self):
        refset = [
            ["V1", "label1", (6.14194444444, 48.67)],
            ["V2", "label2", (6.2, 49)],
            ["V3", "label3", (5.1, 48)],
            ["V4", "label4", (5.2, 48.1)],
        ]
        targetset = [
            ["T1", "labelt1", (6.17, 48.7)],
            ["T2", "labelt2", (5.3, 48.2)],
            ["T3", "labelt3", (6.25, 48.91)],
        ]
        # Creation of the aligner object
        processings = (GeographicalProcessing(2, 2, units="km"),)
        aligner = alig.BaseAligner(threshold=30, processings=processings)
        mat, matched = aligner.align(refset, targetset)
        true_matched = [(0, 0), (0, 2), (1, 2), (3, 1)]
        for k, values in matched.items():
            for v, distance in values:
                self.assertIn((k, v), true_matched)

    def test_align_with_weight(self):
        refset = [
            ["V1", "l1", (6.14194444444, 48.67)],
            ["V2", "l2", (6.2, 49)],
            ["V3", "l3", (5.1, 48)],
            ["V4", "l4", (5.2, 48.1)],
        ]
        targetset = [
            ["T1", "l1", (6.17, 48.7)],
            ["T2", "l1", (6.14, 48.73)],
            ["T3", "l4", (6.25, 48.91)],
        ]

        ref_indexes = range(len(refset))
        target_indexes = range(len(targetset))

        # Compute the distance matrix on the label only (weight=1)
        aligner = alig.BaseAligner(
            threshold=0.2,
            processings=(LevenshteinProcessing(1, 1),),
        )
        mat_difflib = aligner.compute_distance_matrix(
            refset,
            targetset,
            ref_indexes,
            target_indexes,
        )

        # Compute the distance matrix on the location only (weight=1)
        aligner = alig.BaseAligner(threshold=0.2, processings=(GeographicalProcessing(2, 2),))
        mat_geo = aligner.compute_distance_matrix(
            refset,
            targetset,
            ref_indexes,
            target_indexes,
        )

        # Compute the distance matrix with label and location (same weight)
        aligner = alig.BaseAligner(
            threshold=0.2,
            processings=(
                LevenshteinProcessing(1, 1),
                GeographicalProcessing(2, 2),
            ),
        )
        mat_difflib_geo = aligner.compute_distance_matrix(
            refset,
            targetset,
            ref_indexes,
            target_indexes,
        )

        np.testing.assert_almost_equal(mat_difflib_geo, mat_geo + mat_difflib)

        # Same aligner, but the processings have ≠ weights, to show that the
        # weight is taken into account in the final result
        aligner = alig.BaseAligner(
            threshold=0.2,
            processings=(
                LevenshteinProcessing(1, 1, weight=0.8),
                GeographicalProcessing(2, 2, weight=0.2),
            ),
        )
        mat_difflib_geo = aligner.compute_distance_matrix(
            refset,
            targetset,
            ref_indexes,
            target_indexes,
        )

        np.testing.assert_almost_equal(mat_difflib_geo, 0.2 * mat_geo + 0.8 * mat_difflib)

    def test_blocking_align(self):
        refset = [
            ["V1", "label1", (6.14194444444, 48.67)],
            ["V2", "label2", (6.2, 49)],
            ["V3", "label3", (5.1, 48)],
            ["V4", "label4", (5.2, 48.1)],
        ]
        targetset = [
            ["T1", "labelt1", (6.17, 48.7)],
            ["T2", "labelt2", (5.3, 48.2)],
            ["T3", "labelt3", (6.25, 48.91)],
        ]
        # Creation of the aligner object
        true_matched = {(0, 0), (0, 2), (1, 2), (3, 1)}
        processings = (GeographicalProcessing(2, 2, units="km"),)
        aligner = alig.BaseAligner(threshold=30, processings=processings)
        blocking = blo.KdTreeBlocking(ref_attr_index=2, target_attr_index=2, threshold=0.3)
        blocking.fit(refset, targetset)
        predict_matched = set()
        for alignind, targetind in blocking.iter_indice_blocks():
            mat, matched = aligner._get_match(refset, targetset, alignind, targetind)
            for k, values in matched.items():
                for v, distance in values:
                    predict_matched.add((k, v))
        self.assertEqual(true_matched, predict_matched)

    def test_blocking_align_2(self):
        refset = [
            ["V1", "label1", (6.14194444444, 48.67)],
            ["V2", "label2", (6.2, 49)],
            ["V3", "label3", (5.1, 48)],
            ["V4", "label4", (5.2, 48.1)],
        ]
        targetset = [
            ["T1", "labelt1", (6.17, 48.7)],
            ["T2", "labelt2", (5.3, 48.2)],
            ["T3", "labelt3", (6.25, 48.91)],
        ]
        # Creation of the aligner object
        true_matched = {(0, 0), (0, 2), (1, 2), (3, 1)}
        processings = (GeographicalProcessing(2, 2, units="km"),)
        aligner = alig.BaseAligner(threshold=30, processings=processings)
        aligner.register_blocking(
            blo.KdTreeBlocking(ref_attr_index=2, target_attr_index=2, threshold=0.3)
        )
        global_mat, global_matched = aligner.align(refset, targetset)
        predict_matched = set()
        for k, values in global_matched.items():
            for v, distance in values:
                predict_matched.add((k, v))
        self.assertEqual(true_matched, predict_matched)

    def test_unique_align(self):
        refset = [
            ["V1", "label1", (6.14194444444, 48.67)],
            ["V2", "label2", (6.2, 49)],
            ["V3", "label3", (5.1, 48)],
            ["V4", "label4", (5.2, 48.1)],
        ]
        targetset = [
            ["T1", "labelt1", (6.17, 48.7)],
            ["T2", "labelt2", (5.3, 48.2)],
            ["T3", "labelt3", (6.25, 48.91)],
        ]
        all_matched = [
            (("V1", 0), ("T3", 2)),
            (("V1", 0), ("T1", 0)),
            (("V2", 1), ("T3", 2)),
            (("V4", 3), ("T2", 1)),
        ]
        uniq_matched = [(("V1", 0), ("T1", 0)), (("V2", 1), ("T3", 2)), (("V4", 3), ("T2", 1))]
        processings = (GeographicalProcessing(2, 2, units="km"),)
        aligner = alig.BaseAligner(threshold=30, processings=processings)
        aligner.register_blocking(
            blo.KdTreeBlocking(ref_attr_index=2, target_attr_index=2, threshold=0.3)
        )
        unimatched = list(aligner.get_aligned_pairs(refset, targetset, unique=True))
        unimatched_wo_distance = [r[:2] for r in unimatched]
        matched = list(aligner.get_aligned_pairs(refset, targetset, unique=False))
        matched_wo_distance = [r[:2] for r in matched]
        self.assertEqual(len(matched), len(all_matched))
        for m in all_matched:
            self.assertIn(m, matched_wo_distance)
        self.assertEqual(len(unimatched), len(uniq_matched))
        for m in uniq_matched:
            self.assertIn(m, unimatched_wo_distance)

    def test_align_from_file(self):
        uniq_matched = [(("V1", 0), ("T1", 0)), (("V2", 1), ("T3", 2)), (("V4", 3), ("T2", 1))]
        processings = (GeographicalProcessing(2, 2, units="km"),)
        aligner = alig.BaseAligner(threshold=30, processings=processings)
        aligner.register_blocking(
            blo.KdTreeBlocking(ref_attr_index=2, target_attr_index=2, threshold=0.3)
        )
        matched = list(
            aligner.get_aligned_pairs_from_files(
                path.join(TESTDIR, "data", "alignfile.csv"),
                path.join(TESTDIR, "data", "targetfile.csv"),
                ref_indexes=[0, 1, (2, 3)],
                target_indexes=[0, 1, (2, 3)],
            )
        )
        matched_wo_distance = [r[:2] for r in matched]
        self.assertEqual(len(matched), len(uniq_matched))
        for m in uniq_matched:
            self.assertIn(m, matched_wo_distance)


class PipelineAlignerTestCase(unittest.TestCase):
    def test_pipeline_align_pairs(self):
        refset = [
            ["V1", "aaa", (6.14194444444, 48.67)],
            ["V2", "bbb", (6.2, 49)],
            ["V3", "ccc", (5.1, 48)],
            ["V4", "ddd", (5.2, 48.1)],
        ]
        targetset = [
            ["T1", "zzz", (6.17, 48.7)],
            ["T2", "eec", (5.3, 48.2)],
            ["T3", "fff", (6.25, 48.91)],
            ["T4", "ccd", (0, 0)],
        ]
        # Creation of the aligner object
        processings = (GeographicalProcessing(2, 2, units="km"),)
        aligner_1 = alig.BaseAligner(threshold=30, processings=processings)
        processings = (LevenshteinProcessing(1, 1),)
        aligner_2 = alig.BaseAligner(threshold=1, processings=processings)
        pipeline = alig.PipelineAligner((aligner_1, aligner_2))
        uniq_matched = [
            (("V1", 0), ("T1", 0)),
            (("V2", 1), ("T3", 2)),
            (("V4", 3), ("T2", 1)),
            (("V3", 2), ("T4", 3)),
        ]
        matched = list(pipeline.get_aligned_pairs(refset, targetset, unique=True))
        matched_wo_distance = [r[:2] for r in matched]
        self.assertEqual(len(matched), len(uniq_matched))
        for m in uniq_matched:
            self.assertIn(m, matched_wo_distance)

        expected_distances = [1.00000, 4.5532517, 11.396689, 15.692410]
        for index, (_, _, distance) in enumerate(sorted(matched, key=lambda x: x[2])):
            self.assertAlmostEqual(distance, expected_distances[index], 6)


if __name__ == "__main__":
    unittest.main()



#
# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest
import shutil
from contextlib import contextmanager
from os import path
from tempfile import mkdtemp

from nazca.utils.dataio import (
    HTMLPrettyPrint,
    ValidXHTMLPrettyPrint,
    sparqlquery,
    sparqljson,
    _sparqlexecute,
    parsefile,
    autocast,
    split_file,
    SPARQL_ENABLED,
)
from nazca.utils.tokenizer import NLTK_AVAILABLE
from nazca.ner import NerProcess
from nazca.ner.sources import NerSourceLexicon

TESTDIR = path.dirname(__file__)


@contextmanager
def tempdir():
    try:
        temp = mkdtemp()
        yield temp
    finally:
        try:
            shutil.rmtree(temp)
        except BaseException:
            pass


class ValidXHTMLPrettyPrintTest(unittest.TestCase):
    def test_valid(self):
        from lxml import etree

        if int(etree.__version__ < "3.2.0"):
            # https://bugs.launchpad.net/lxml/+bug/673205
            self.skipTest("Lxml version to old for ValidXHTMLPrettyPrint")
        self.assertTrue(ValidXHTMLPrettyPrint().is_valid("<p>coucou</p>"))

    def test_valid_unicode(self):
        from lxml import etree

        if int(etree.__version__ < "3.2.0"):
            # https://bugs.launchpad.net/lxml/+bug/673205
            self.skipTest("Lxml version to old for ValidXHTMLPrettyPrint")
        self.assertTrue(ValidXHTMLPrettyPrint().is_valid("<p>hé</p>"))

    def test_invalid(self):
        from lxml import etree

        if int(etree.__version__ < "3.2.0"):
            # https://bugs.launchpad.net/lxml/+bug/673205
            self.skipTest("Lxml version to old for ValidXHTMLPrettyPrint")
        self.assertFalse(ValidXHTMLPrettyPrint().is_valid("<p><div>coucou</div></p>"))

    @unittest.skipUnless(NLTK_AVAILABLE, "nltk is not available")
    def test_prettyprint(self):
        text = "Hello everyone, this is   me speaking. And me."
        source = NerSourceLexicon(
            {"everyone": "http://example.com/everyone", "me": "http://example.com/me"}
        )
        ner = NerProcess((source,))
        named_entities = ner.process_text(text)
        html = HTMLPrettyPrint().pprint_text(text, named_entities)
        self.assertEqual(
            html,
            (
                'Hello <a href="http://example.com/everyone">everyone</a>, '
                'this is   <a href="http://example.com/me">me</a> speaking. '
                'And <a href="http://example.com/me">me</a>.'
            ),
        )

    @unittest.skipUnless(NLTK_AVAILABLE, "nltk is not available")
    def test_prettyprint_class(self):
        text = "Hello everyone, this is   me speaking. And me."
        source = NerSourceLexicon(
            {"everyone": "http://example.com/everyone", "me": "http://example.com/me"}
        )
        ner = NerProcess((source,))
        named_entities = ner.process_text(text)
        html = HTMLPrettyPrint().pprint_text(text, named_entities, html_class="ner")
        self.assertEqual(
            html,
            (
                'Hello <a href="http://example.com/everyone" class="ner">everyone</a>, '
                'this is   <a href="http://example.com/me" class="ner">me</a> speaking. '
                'And <a href="http://example.com/me" class="ner">me</a>.'
            ),
        )


class DataIOTestCase(unittest.TestCase):
    def test_parser(self):
        data = parsefile(path.join(TESTDIR, "data", "file2parse"), [0, (2, 3), 4, 1], delimiter=",")
        self.assertEqual(
            [
                [1, (12, 19), "apple", "house"],
                [2, (21.9, 19), "stramberry", "horse"],
                [3, (23, 2.17), "cherry", "flower"],
            ],
            data,
        )

        data = parsefile(
            path.join(TESTDIR, "data", "file2parse"),
            [0, (2, 3), 4, 1],
            delimiter=",",
            formatopt={2: str},
        )
        self.assertEqual(
            [
                [1, ("12", 19), "apple", "house"],
                [2, ("21.9", 19), "stramberry", "horse"],
                [3, ("23", 2.17), "cherry", "flower"],
            ],
            data,
        )

        data = parsefile(
            path.join(TESTDIR, "data", "file2parse"),
            [0, (2, 3), 4, 1],
            delimiter=",",
            autocast_data=False,
        )
        self.assertEqual(
            [
                ["1", ("12", "19"), "apple", "house"],
                ["2", ("21.9", "19"), "stramberry", "horse"],
                ["3", ("23", "2.17"), "cherry", "flower"],
            ],
            data,
        )

    def test_autocast(self):
        self.assertEqual(autocast("1"), 1)
        self.assertEqual(autocast("1."), 1.0)
        self.assertEqual(autocast("1,"), 1.0)
        self.assertEqual(autocast("1,2"), 1.2)
        self.assertEqual(autocast("1,2X"), "1,2X")
        self.assertEqual(autocast("tété"), "tété")
        self.assertEqual(autocast("tété", encoding="utf-8"), "tété")

    def test_split_file(self):
        NBLINES = 190
        with tempdir() as outputdir:
            file2split = path.join(TESTDIR, "data", "file2split")
            files = split_file(file2split, outputdir, nblines=NBLINES)

            alllines = []
            nbfiles = len(files)
            for num, localpath in enumerate(sorted(files)):
                fullpath = path.join(outputdir, localpath)
                with open(fullpath) as fobj:
                    lines = fobj.readlines()
                # All files, except the last one, must be NBLINES-length.
                if num < nbfiles - 1:
                    self.assertEqual(len(lines), NBLINES)
                alllines.extend(lines)

            with open(file2split) as fobj:
                self.assertEqual(alllines, fobj.readlines())

    @unittest.skipUnless(SPARQL_ENABLED, "python-sparqlwrapper is not installed")
    def test_sparql_query(self):
        results = sparqlquery(
            "http://dbpedia.org/sparql",
            """\
            PREFIX dbpedia-owl:<http://dbpedia.org/ontology/>
            SELECT DISTINCT ?uri ?designer
            WHERE {
                ?uri rdf:type <http://dbpedia.org/ontology/ProgrammingLanguage> ;
                rdfs:label ?label ;
                dbpedia-owl:designer ?designer .
            FILTER(regex(?label, "^Python"))
            }""",
        )
        self.assertEqual(
            results,
            [
                [
                    "http://dbpedia.org/resource/Python_(programming_language)",
                    "http://dbpedia.org/resource/Guido_van_Rossum",
                ],
                [
                    "http://dbpedia.org/resource/Python_for_S60",
                    "http://dbpedia.org/resource/Guido_van_Rossum",
                ],
            ],
        )

    @unittest.skipUnless(SPARQL_ENABLED, "python-sparqlwrapper is not installed")
    def test_sparql_execute(self):
        rawresults = _sparqlexecute(
            "http://dbpedia.org/sparql",
            """\
            PREFIX dbpedia-owl:<http://dbpedia.org/ontology/>
            SELECT DISTINCT ?uri ?designer
            WHERE {
            ?uri rdf:type <http://dbpedia.org/ontology/ProgrammingLanguage> ;
                 rdfs:label ?label ;
                 dbpedia-owl:designer ?designer .
            FILTER(regex(?label, "^Python"))
            }""",
        )
        self.assertEqual(
            rawresults,
            {
                "head": {"link": [], "vars": ["uri", "designer"]},
                "results": {
                    "bindings": [
                        {
                            "designer": {
                                "type": "uri",
                                "value": "http://dbpedia.org/resource/Guido_van_Rossum",
                            },
                            "uri": {
                                "type": "uri",
                                "value": (
                                    "http://dbpedia.org/resource/" "Python_(programming_language)"
                                ),
                            },
                        },
                        {
                            "designer": {
                                "type": "uri",
                                "value": "http://dbpedia.org/resource/Guido_van_Rossum",
                            },
                            "uri": {
                                "type": "uri",
                                "value": "http://dbpedia.org/resource/Python_for_S60",
                            },
                        },
                    ],
                    "distinct": False,
                    "ordered": True,
                },
            },
        )

    @unittest.skipUnless(SPARQL_ENABLED, "python-sparqlwrapper is not installed")
    def test_sparql_execute_no_raise_on_error(self):
        rawresults = _sparqlexecute(
            "http://dbpedia.org/sparql",
            """SELECT DISTINCT ?uri
                                    WHERE{
                                    ?uri faultyrdf
                                    ?uri rdf:type ?type}""",
        )
        self.assertEqual(rawresults, [])

    @unittest.skipUnless(SPARQL_ENABLED, "python-sparqlwrapper is not installed")
    def test_sparql_execute_raise_on_error(self):
        with self.assertRaises(RuntimeError):
            _sparqlexecute(
                "http://dbpedia.org/sparql",
                """SELECT DISTINCT ?uri
                                        WHERE{
                                        ?uri faultyrdf
                                        ?uri rdf:type ?type}""",
                raise_on_error=True,
            )

    @unittest.skipUnless(SPARQL_ENABLED, "python-sparqlwrapper is not installed")
    def test_sparql_json(self):
        results = sparqljson(
            "http://dbpedia.org/sparql",
            """\
            PREFIX dbpedia-owl:<http://dbpedia.org/ontology/>
            SELECT DISTINCT ?uri ?designer
            WHERE {
            ?uri rdf:type <http://dbpedia.org/ontology/ProgrammingLanguage> ;
                 rdfs:label ?label ;
                 dbpedia-owl:designer ?designer .
            FILTER(regex(?label, "^Python"))
            }""",
        )
        self.assertEqual(
            results,
            {
                "designer": {"http://dbpedia.org/resource/Guido_van_Rossum"},
                "uri": {
                    "http://dbpedia.org/resource/Python_(programming_language)",
                    "http://dbpedia.org/resource/Python_for_S60",
                },
            },
        )

    @unittest.skipUnless(SPARQL_ENABLED, "python-sparqlwrapper is not installed")
    def test_sparql_autocast(self):
        alignset = sparqlquery(
            "http://dbpedia.inria.fr/sparql",
            "prefix db-owl: <http://dbpedia.org/ontology/>"
            "prefix db-prop: <http://fr.dbpedia.org/property/>"
            "select ?ville, ?name, ?long, ?lat where {"
            " ?ville db-owl:country <http://fr.dbpedia.org/resource/France> ."
            " ?ville rdf:type db-owl:PopulatedPlace ."
            " ?ville db-owl:populationTotal ?population ."
            " ?ville foaf:name ?name ."
            " ?ville db-prop:longitude ?long ."
            " ?ville db-prop:latitude ?lat ."
            " FILTER (?population > 1000)"
            "} LIMIT 100",
            indexes=[0, 1, (2, 3)],
        )
        self.assertEqual(len(alignset), 100)
        self.assertTrue(isinstance(alignset[0][2][0], float))

    @unittest.skipUnless(SPARQL_ENABLED, "python-sparqlwrapper is not installed")
    def test_sparql_no_autocast(self):
        alignset = sparqlquery(
            "http://dbpedia.inria.fr/sparql",
            "prefix db-owl: <http://dbpedia.org/ontology/>"
            "prefix db-prop: <http://fr.dbpedia.org/property/>"
            "select ?ville, ?name, ?long, ?lat where {"
            " ?ville db-owl:country <http://fr.dbpedia.org/resource/France> ."
            " ?ville rdf:type db-owl:PopulatedPlace ."
            " ?ville db-owl:populationTotal ?population ."
            " ?ville foaf:name ?name ."
            " ?ville db-prop:longitude ?long ."
            " ?ville db-prop:latitude ?lat ."
            " FILTER (?population > 1000)"
            "} LIMIT 100",
            indexes=[0, 1, (2, 3)],
            autocast_data=False,
        )
        self.assertEqual(len(alignset), 100)
        self.assertFalse(isinstance(alignset[0][2][0], float))


if __name__ == "__main__":
    unittest.main()



# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import re
from string import punctuation
from unicodedata import normalize as _uninormalize
from functools import partial

FRENCH_STOPWORDS = {
    "alors",
    "au",
    "aux",
    "aucuns",
    "aussi",
    "autre",
    "avant",
    "avec",
    "avoir",
    "bon",
    "car",
    "ce",
    "cela",
    "ces",
    "ceux",
    "chaque",
    "ci",
    "comme",
    "comment",
    "dans",
    "de",
    "des",
    "du",
    "dedans",
    "dehors",
    "depuis",
    "deux",
    "devrait",
    "doit",
    "donc",
    "dos",
    "droite",
    "début",
    "elle",
    "elles",
    "en",
    "encore",
    "essai",
    "est",
    "et",
    "eu",
    "eux",
    "fait",
    "faites",
    "fois",
    "font",
    "force",
    "haut",
    "hors",
    "ici",
    "il",
    "ils",
    "je",
    "juste",
    "la",
    "le",
    "les",
    "leur",
    "lui",
    "là",
    "ma",
    "maintenant",
    "mais",
    "me",
    "mes",
    "moi",
    "moins",
    "mon",
    "mot",
    "même",
    "ne",
    "ni",
    "nommés",
    "nos",
    "notre",
    "nous",
    "nouveaux",
    "on",
    "ou",
    "où",
    "par",
    "parce",
    "parole",
    "pas",
    "personnes",
    "peut",
    "peu",
    "pièce",
    "plupart",
    "pour",
    "pourquoi",
    "quand",
    "que",
    "quel",
    "quelle",
    "quelles",
    "quels",
    "qui",
    "sa",
    "sans",
    "se",
    "ses",
    "seulement",
    "si",
    "sien",
    "son",
    "sont",
    "sous",
    "soyez",
    "sujet",
    "sur",
    "ta",
    "tandis",
    "tellement",
    "te",
    "tels",
    "tes",
    "toi",
    "ton",
    "tous",
    "tout",
    "trop",
    "très",
    "tu",
    "un",
    "une",
    "valeur",
    "voie",
    "voient",
    "vont",
    "vos",
    "votre",
    "vous",
    "vu",
    "ça",
    "étaient",
    "état",
    "étions",
    "été",
    "être",
}

MANUAL_UNICODE_MAP = {
    "\xa1": "!",  # INVERTED EXCLAMATION MARK
    "\u0142": "l",  # LATIN SMALL LETTER L WITH STROKE
    "\u2044": "/",  # FRACTION SLASH
    "\xc6": "AE",  # LATIN CAPITAL LETTER AE
    "\xa9": "(c)",  # COPYRIGHT SIGN
    "\xab": '"',  # LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    "\xe6": "ae",  # LATIN SMALL LETTER AE
    "\xae": "(r)",  # REGISTERED SIGN
    "\u0153": "oe",  # LATIN SMALL LIGATURE OE
    "\u0152": "OE",  # LATIN CAPITAL LIGATURE OE
    "\xd8": "O",  # LATIN CAPITAL LETTER O WITH STROKE
    "\xf8": "o",  # LATIN SMALL LETTER O WITH STROKE
    "\xbb": '"',  # RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    "\xdf": "ss",  # LATIN SMALL LETTER SHARP S
    "\u2019": "'",  # RIGHT SINGLE QUOTATION MARK
}


###############################################################################
# NORMALIZE FUNCTIONS #########################################################
###############################################################################
def unormalize(ustring, substitute=None):
    """replace diacritical characters with their corresponding ascii characters

    Convert the unicode string to its long normalized form (unicode character
    will be transform into several characters) and keep the first one only.
    The normal form KD (NFKD) will apply the compatibility decomposition, i.e.
    replace all compatibility characters with their equivalents.

    :type substitute: str
    :param substitute: replacement character to use if decomposition fails

    :see: Another project about ASCII transliterations of Unicode text
          http://pypi.python.org/pypi/Unidecode
    """
    res = []
    for letter in ustring[:]:
        try:
            replacement = MANUAL_UNICODE_MAP[letter]
        except KeyError:
            if isinstance(letter, str):
                replacement = _uninormalize("NFKD", letter)[0]
            else:
                replacement = letter
            if ord(replacement) >= 2**7:
                if substitute is None:
                    raise ValueError("can't deal with non-ascii based characters")
                replacement = substitute
        res.append(replacement)
    return "".join(res)


def lunormalize(sentence, substitute=None):
    """Normalize a sentence (ie remove accents, set to lower, etc)"""
    return unormalize(sentence, substitute).lower()


def simplify(sentence, lemmas=None, remove_stopwords=True, stopwords=FRENCH_STOPWORDS):
    """Simply the given sentence
    0) If remove_stopwords, then remove the stop words
    1) If lemmas are given, the sentence is lemmatized
    2) Set the sentence to lower case
    3) Remove punctuation
    """
    if not isinstance(sentence, str):
        return sentence

    if lemmas:
        sentence = lemmatized(sentence, lemmas)
    sentence = sentence.lower()
    cleansent = "".join([s if s not in punctuation else " " for s in sentence]).strip()
    # comma followed by a space is replaced by two spaces, keep only one
    cleansent = cleansent.replace("  ", " ")

    if not remove_stopwords:
        return cleansent
    else:
        return " ".join([w for w in cleansent.split(" ") if w not in stopwords])


def tokenize(sentence, tokenizer=None, regexp=re.compile(r"[^\s]+")):
    """Tokenize a sentence.
    Use ``tokenizer`` if given, else try to use the nltk WordPunctTokenizer,
    in case of failure, it just split on spaces.

    Anyway, tokenizer must have a ``tokenize()`` method
    """
    if tokenizer:
        return tokenizer().tokenize(sentence)
    # XXX Unicode, could not use WorkTokenizer.
    # Instead split on whitespaces
    chunks = []
    for chunk in (t for t in regexp.findall(sentence) if t):
        # Deals with '
        if "'" in chunk:
            schunks = chunk.split("'")
            chunks.extend([c + "'" for c in schunks[:-1]])
            chunks.append(schunks[-1])
        else:
            chunks.append(chunk)
    return chunks


def lemmatized(sentence, lemmas, tokenizer=None):
    """Return the lemmatized sentence"""
    tokenized_sent = tokenize(sentence, tokenizer)
    tokenized_sentformated = []
    for w in tokenized_sent:
        if w in ".,'" and len(tokenized_sentformated) > 0:
            tokenized_sentformated[-1] += w
        elif w not in punctuation:
            tokenized_sentformated.append(w)
    return " ".join([lemmatized_word(w, lemmas) for w in tokenized_sentformated])


def lemmatized_word(word, lemmas):
    """Return the lemmatized word"""
    lemma = lemmas.get(word.lower(), word)
    if "|" in lemma:
        _words = lemma.split("|")
        if word.lower() in _words:
            lemma = word.lower()
        else:
            lemma = _words[0]
    return lemma


def roundstr(number, ndigits=0):
    """Return an unicode string of ``number`` rounded to a given precision
    in decimal digits (default 0 digits)

    If ``number`` is not a float, this method casts it to a float. (An
    exception may be raised if it's not possible)
    """
    return format(round(float(number), ndigits), "0.%df" % ndigits)


def rgxformat(string, regexp, output):
    r"""Apply the regexp to the ``string`` and return a formatted string
    according to ``output``

    eg :
        format(u'[Victor Hugo - 26 fev 1802 / 22 mai 1885]',
               r'\[(?P<firstname>\w+) (?p<lastname>\w+) - '
               r'(?P<birthdate>.*?) / (?<deathdate>.*?)\]',
               u'%(lastname)s, %(firstname)s (%(birthdate)s -'
               u'%(deathdate)s)')

     would return u'Hugo, Victor (26 fev 1802 - 22 mai 1885)'
    """
    match = re.match(regexp, string)
    return output % match.groupdict()


###############################################################################
# NORMALIZER OBJECTS ##########################################################
###############################################################################
class BaseNormalizer:
    """A normalizer object used to provide an abstraction over the different
    normalization functions, and help building Nazca process."""

    def __init__(self, callback, attr_index=None):
        """Initiate the BaseNormalizer

        Parameters
        ----------
        callback: normalization callback

        attr_index: index of the attribute of interest in a record
                    (i.e. attribute to be normalized).
                    By default, 'attr_index' is None and the whole
                    record is passed to the callback.
                    If given, only the attr_index value of the record
                    is passed to the the callback.
                    Could be a list or an int
        """
        self.callback = callback
        if attr_index is not None:
            self.attr_index = attr_index if isinstance(attr_index, (tuple, list)) else (attr_index,)
        else:
            self.attr_index = attr_index

    def normalize(self, record):
        """Normalize a record

        Parameters
        ----------
        record: a record (tuple/list of values).

        Returns
        -------

        record: the normalized record.
        """
        if not self.attr_index:
            return self.callback(record)
        else:
            for attr_ind in self.attr_index:
                record = list(
                    r if ind != attr_ind else self.callback(r) for ind, r in enumerate(record)
                )
            return record

    def normalize_dataset(self, dataset, inplace=False):
        """Normalize a dataset

        Parameters
        ----------
        dataset: a list of record (tuple/list of values).

        inplace: Boolean. If True, normalize the dataset in place.

        Returns
        -------

        record: the normalized dataset.
        """
        if not inplace:
            dataset = [self.normalize(record) for record in dataset]
        else:
            # Change dataset in place
            for ind, record in enumerate(dataset):
                dataset[ind] = self.normalize(record)
        return dataset


class UnicodeNormalizer(BaseNormalizer):
    """Normalizer that unormalize the unicode
    (i.e. replace accentuating characters by ASCII ones)
    """

    def __init__(self, attr_index=None, substitute=None):
        callback = partial(lunormalize, substitute=substitute)
        super().__init__(callback, attr_index=attr_index)


class SimplifyNormalizer(BaseNormalizer):
    """Normalizer that simplify a string
    0) If remove_stopwords, then remove the stop words
    1) If lemmas are given, the sentence is lemmatized
    2) Set the sentence to lower case
    3) Remove punctuation
    """

    def __init__(self, attr_index=None, lemmas=None, remove_stopwords=True):
        callback = partial(simplify, lemmas=lemmas, remove_stopwords=remove_stopwords)
        super().__init__(callback, attr_index=attr_index)


class TokenizerNormalizer(BaseNormalizer):
    """Normalizer that tokenize a string
    Use ``tokenizer`` if given, else try to use the nltk WordPunctTokenizer,
    in case of failure, it just split on spaces.
    Anyway, tokenizer must have a ``tokenize()`` method
    """

    def __init__(self, attr_index=None, tokenizer=None, regexp=re.compile(r"[^\s]+")):
        callback = partial(tokenize, tokenizer=tokenizer, regexp=regexp)
        super().__init__(callback, attr_index=attr_index)


class LemmatizerNormalizer(BaseNormalizer):
    """Normalizer that lemmatize a string"""

    def __init__(self, lemmas, attr_index=None, tokenizer=None):
        callback = partial(lemmatized, lemmas=lemmas, tokenizer=tokenizer)
        super().__init__(callback, attr_index=attr_index)


class RoundNormalizer(BaseNormalizer):
    """Normalizer that round a string
    Return an unicode string of ``number`` rounded to a given precision
    in decimal digits (default 0 digits)

    If ``number`` is not a float, this method casts it to a float. (An
    exception may be raised if it's not possible)
    """

    def __init__(self, attr_index=None, ndigits=0):
        callback = partial(roundstr, ndigits=ndigits)
        super().__init__(callback, attr_index=attr_index)


class RegexpNormalizer(BaseNormalizer):
    r"""Normalizer that normalize a string based on a regexp

     Apply the regexp to the ``string`` and return a formatted string
    according to ``output``

    eg :
        format(u'[Victor Hugo - 26 fev 1802 / 22 mai 1885]',
               r'\[(?P<firstname>\w+) (?p<lastname>\w+) - '
               r'(?P<birthdate>.*?) / (?<deathdate>.*?)\]',
               u'%(lastname)s, %(firstname)s (%(birthdate)s -'
               u'%(deathdate)s)')

     would return u'Hugo, Victor (26 fev 1802 - 22 mai 1885)'
    """

    def __init__(self, regexp, output, attr_index=None):
        callback = partial(rgxformat, regexp=regexp, output=output)
        super().__init__(callback, attr_index=attr_index)


###############################################################################
# JOIN NORMALIZER #############################################################
###############################################################################
class JoinNormalizer(BaseNormalizer):
    """Normalizer that join multiple fields in only one.
    This new field will be put at the end of the new record.
    """

    def __init__(self, attr_indexes, join_car=", "):
        self.attr_indexes = attr_indexes
        self.join_car = join_car

    def normalize(self, record):
        """Normalize a record

        Parameters
        ----------
        record: a record (tuple/list of values).

        Returns
        -------

        record: the normalized record.
        """
        _record = [r for ind, r in enumerate(record) if ind not in self.attr_indexes]
        _record.append(
            self.join_car.join([r for ind, r in enumerate(record) if ind in self.attr_indexes])
        )
        return _record


###############################################################################
# NORMALIZER PIPELINE #########################################################
###############################################################################
class NormalizerPipeline(BaseNormalizer):
    """Pipeline of Normalizers"""

    def __init__(self, normalizers):
        """Initiate the NormalizerPipeline

        Parameters
        ----------
        normalizers: list (ordered) of Normalizer
        """
        self.normalizers = normalizers

    def normalize(self, record):
        """Normalize a record

        Parameters
        ----------
        record: a record (tuple/list of values).

        Returns
        -------

        record: the normalized record.
        """
        for normalizer in self.normalizers:
            record = normalizer.normalize(record)
        return record



#
# copyright 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

from nazca.ner.sources import NerSourceLexicon, NerSourceSparql
from nazca.ner import NerProcess
from nazca.utils.tokenizer import Token, Sentence, NLTK_AVAILABLE
from nazca.ner.preprocessors import NerStopwordsFilterPreprocessor
from nazca.utils.dataio import SPARQL_ENABLED


class NerTest(unittest.TestCase):
    """Test of Ner"""

    def test_lexicon_source(self):
        """Test lexicon source"""
        lexicon = {"everyone": "http://example.com/everyone", "me": "http://example.com/me"}
        source = NerSourceLexicon(lexicon)
        self.assertEqual(
            source.query_word("me"),
            [
                "http://example.com/me",
            ],
        )
        self.assertEqual(
            source.query_word("everyone"),
            [
                "http://example.com/everyone",
            ],
        )
        self.assertEqual(source.query_word("me everyone"), [])
        self.assertEqual(source.query_word("toto"), [])
        # Token
        token = Token("me", 0, 2, None)
        self.assertEqual(
            source.recognize_token(token),
            [
                "http://example.com/me",
            ],
        )
        token = Token("ma", 0, 2, None)
        self.assertEqual(source.recognize_token(token), [])

    @unittest.skipUnless(SPARQL_ENABLED, "python-sparqlwrapper is not installed")
    def test_sparql_source(self):
        """Test sparql source"""
        source = NerSourceSparql(
            "http://dbpedia.org/sparql",
            """\
            PREFIX dbpedia-owl:<http://dbpedia.org/ontology/>
            SELECT DISTINCT ?uri
            WHERE {
            ?uri rdf:type <http://dbpedia.org/ontology/ProgrammingLanguage> ;
                 dbpedia-owl:designer <http://dbpedia.org/resource/Guido_van_Rossum> ;
                 rdfs:label ?label.
            FILTER(regex(?label, "^%(word)s"))
            }""",
        )
        self.assertEqual(
            source.query_word("Python"),
            [
                "http://dbpedia.org/resource/Python_(programming_language)",
                "http://dbpedia.org/resource/Python_for_S60",
            ],
        )

    @unittest.skipUnless(NLTK_AVAILABLE, "nltk is not available")
    def test_ner_process(self):
        """Test ner process"""
        text = "Hello everyone, this is   me speaking. And me."
        source = NerSourceLexicon(
            {"everyone": "http://example.com/everyone", "me": "http://example.com/me"}
        )
        ner = NerProcess((source,))
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/everyone",
                    None,
                    Token(
                        word="everyone",
                        start=6,
                        end=14,
                        sentence=Sentence(indice=0, start=0, end=38),
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=26, end=28, sentence=Sentence(indice=0, start=0, end=38)
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=43, end=45, sentence=Sentence(indice=1, start=39, end=46)
                    ),
                ),
            ],
        )

    @unittest.skipUnless(NLTK_AVAILABLE, "nltk is not available")
    def test_ner_process_multisources(self):
        """Test ner process"""
        text = "Hello everyone, this is   me speaking. And me."
        source1 = NerSourceLexicon(
            {"everyone": "http://example.com/everyone", "me": "http://example.com/me"}
        )
        source2 = NerSourceLexicon({"me": "http://example2.com/me"})
        # Two sources, not unique
        ner = NerProcess((source1, source2))
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/everyone",
                    None,
                    Token(
                        word="everyone",
                        start=6,
                        end=14,
                        sentence=Sentence(indice=0, start=0, end=38),
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=26, end=28, sentence=Sentence(indice=0, start=0, end=38)
                    ),
                ),
                (
                    "http://example2.com/me",
                    None,
                    Token(
                        word="me", start=26, end=28, sentence=Sentence(indice=0, start=0, end=38)
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=43, end=45, sentence=Sentence(indice=1, start=39, end=46)
                    ),
                ),
                (
                    "http://example2.com/me",
                    None,
                    Token(
                        word="me", start=43, end=45, sentence=Sentence(indice=1, start=39, end=46)
                    ),
                ),
            ],
        )
        # Two sources, unique
        ner = NerProcess((source1, source2), unique=True)
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/everyone",
                    None,
                    Token(
                        word="everyone",
                        start=6,
                        end=14,
                        sentence=Sentence(indice=0, start=0, end=38),
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=26, end=28, sentence=Sentence(indice=0, start=0, end=38)
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=43, end=45, sentence=Sentence(indice=1, start=39, end=46)
                    ),
                ),
            ],
        )
        # Two sources inversed, unique
        ner = NerProcess((source2, source1), unique=True)
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/everyone",
                    None,
                    Token(
                        word="everyone",
                        start=6,
                        end=14,
                        sentence=Sentence(indice=0, start=0, end=38),
                    ),
                ),
                (
                    "http://example2.com/me",
                    None,
                    Token(
                        word="me", start=26, end=28, sentence=Sentence(indice=0, start=0, end=38)
                    ),
                ),
                (
                    "http://example2.com/me",
                    None,
                    Token(
                        word="me", start=43, end=45, sentence=Sentence(indice=1, start=39, end=46)
                    ),
                ),
            ],
        )

    @unittest.skipUnless(NLTK_AVAILABLE, "nltk is not available")
    def test_ner_process_add_sources(self):
        """Test ner process"""
        text = "Hello everyone, this is   me speaking. And me."
        source1 = NerSourceLexicon(
            {"everyone": "http://example.com/everyone", "me": "http://example.com/me"}
        )
        source2 = NerSourceLexicon({"me": "http://example2.com/me"})
        ner = NerProcess((source1,))
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/everyone",
                    None,
                    Token(
                        word="everyone",
                        start=6,
                        end=14,
                        sentence=Sentence(indice=0, start=0, end=38),
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=26, end=28, sentence=Sentence(indice=0, start=0, end=38)
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=43, end=45, sentence=Sentence(indice=1, start=39, end=46)
                    ),
                ),
            ],
        )
        # Two sources, not unique
        ner.add_ner_source(source2)
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/everyone",
                    None,
                    Token(
                        word="everyone",
                        start=6,
                        end=14,
                        sentence=Sentence(indice=0, start=0, end=38),
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=26, end=28, sentence=Sentence(indice=0, start=0, end=38)
                    ),
                ),
                (
                    "http://example2.com/me",
                    None,
                    Token(
                        word="me", start=26, end=28, sentence=Sentence(indice=0, start=0, end=38)
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=43, end=45, sentence=Sentence(indice=1, start=39, end=46)
                    ),
                ),
                (
                    "http://example2.com/me",
                    None,
                    Token(
                        word="me", start=43, end=45, sentence=Sentence(indice=1, start=39, end=46)
                    ),
                ),
            ],
        )

    @unittest.skipUnless(NLTK_AVAILABLE, "nltk is not available")
    def test_ner_process_preprocess(self):
        """Test ner process"""
        text = "Hello Toto, this is   me speaking. And me."
        source = NerSourceLexicon(
            {"Toto": "http://example.com/toto", "me": "http://example.com/me"}
        )
        preprocessor = NerStopwordsFilterPreprocessor()
        ner = NerProcess((source,), preprocessors=(preprocessor,))
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/toto",
                    None,
                    Token(
                        word="Toto", start=6, end=10, sentence=Sentence(indice=0, start=0, end=34)
                    ),
                )
            ],
        )

    @unittest.skipUnless(NLTK_AVAILABLE, "nltk is not available")
    def test_ner_process_add_preprocess(self):
        """Test ner process"""
        text = "Hello Toto, this is   me speaking. And me."
        source = NerSourceLexicon(
            {"Toto": "http://example.com/toto", "me": "http://example.com/me"}
        )
        preprocessor = NerStopwordsFilterPreprocessor()
        ner = NerProcess(
            (source,),
        )
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/toto",
                    None,
                    Token(
                        word="Toto", start=6, end=10, sentence=Sentence(indice=0, start=0, end=34)
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=22, end=24, sentence=Sentence(indice=0, start=0, end=34)
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=39, end=41, sentence=Sentence(indice=1, start=35, end=42)
                    ),
                ),
            ],
        )
        ner.add_preprocessors(preprocessor)
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/toto",
                    None,
                    Token(
                        word="Toto", start=6, end=10, sentence=Sentence(indice=0, start=0, end=34)
                    ),
                )
            ],
        )

    @unittest.skipUnless(NLTK_AVAILABLE, "nltk is not available")
    def test_ner_process_chained_word(self):
        """Test ner process"""
        text = "Hello everyone me, this is   me speaking. And me."
        source = NerSourceLexicon(
            {
                "everyone": "http://example.com/everyone",
                "everyone me": "http://example.com/everyone_me",
                "me": "http://example.com/me",
            }
        )
        ner = NerProcess((source,))
        named_entities = ner.process_text(text)
        self.assertEqual(
            named_entities,
            [
                (
                    "http://example.com/everyone_me",
                    None,
                    Token(
                        word="everyone me",
                        start=6,
                        end=17,
                        sentence=Sentence(indice=0, start=0, end=41),
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=29, end=31, sentence=Sentence(indice=0, start=0, end=41)
                    ),
                ),
                (
                    "http://example.com/me",
                    None,
                    Token(
                        word="me", start=46, end=48, sentence=Sentence(indice=1, start=42, end=49)
                    ),
                ),
            ],
        )


if __name__ == "__main__":
    unittest.main()



#
# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest
from os import path

from nazca.utils.normalize import (
    BaseNormalizer,
    UnicodeNormalizer,
    JoinNormalizer,
    SimplifyNormalizer,
    TokenizerNormalizer,
    LemmatizerNormalizer,
    RoundNormalizer,
    RegexpNormalizer,
    NormalizerPipeline,
    lunormalize,
    lemmatized,
    roundstr,
    rgxformat,
    tokenize,
    simplify,
)
from nazca.data import FRENCH_LEMMAS


TESTDIR = path.dirname(__file__)


class NormalizerFunctionTestCase(unittest.TestCase):
    def test_unormalize(self):
        self.assertEqual(lunormalize("bépoèàÀêùï"), "bepoeaaeui")

    def test_simplify(self):
        self.assertEqual(
            simplify("J'aime les frites, les pommes et les" " scoubidous !", FRENCH_LEMMAS),
            "aimer frites pomme scoubidou",
        )

    def test_tokenize(self):
        self.assertEqual(
            tokenize("J'aime les frites !"),
            [
                "J'",
                "aime",
                "les",
                "frites",
                "!",
            ],
        )

    def test_lemmatizer(self):
        self.assertEqual(lemmatized("sacré rubert", FRENCH_LEMMAS), "sacré rubert")
        self.assertEqual(lemmatized("J'aime les frites !", FRENCH_LEMMAS), "je aimer le frite")
        self.assertEqual(lemmatized(", J'aime les frites", FRENCH_LEMMAS), "je aimer le frite")

    def test_round(self):
        self.assertEqual(roundstr(3.14159, 2), "3.14")
        self.assertEqual(roundstr(3.14159), "3")
        self.assertEqual(roundstr("3.14159", 3), "3.142")

    def test_format(self):
        string = "[Victor Hugo - 26 fev 1802 / 22 mai 1885]"
        regex = (
            r"\[(?P<firstname>\w+) (?P<lastname>\w+) - "
            r"(?P<birthdate>.*) \/ (?P<deathdate>.*?)\]"
        )
        output = "%(lastname)s, %(firstname)s (%(birthdate)s - %(deathdate)s)"
        self.assertEqual(
            rgxformat(string, regex, output), "Hugo, Victor (26 fev 1802 - 22 mai 1885)"
        )

        string = "http://perdu.com/42/supertop/cool"
        regex = r"http://perdu.com/(?P<id>\d+).*"
        output = "%(id)s"
        self.assertEqual(rgxformat(string, regex, output), "42")


class NormalizerObjectTestCase(unittest.TestCase):
    def test_normalizer(self):
        normalizer = BaseNormalizer(lunormalize)
        self.assertEqual(normalizer.normalize("bépoèàÀêùï"), "bepoeaaeui")

    def test_normalizer_record(self):
        normalizer = BaseNormalizer(lunormalize, attr_index=1)
        record = ("a1", "bépoèàÀêùï")
        self.assertEqual(normalizer.normalize(record), ["a1", "bepoeaaeui"])

    def test_normalizer_dataset(self):
        normalizer = BaseNormalizer(lunormalize, attr_index=1)
        dataset = [("a1", "bépoèàÀêùï"), ("a2", "tàtà")]
        results = normalizer.normalize_dataset(dataset)
        self.assertEqual([["a1", "bepoeaaeui"], ["a2", "tata"]], results)
        self.assertNotEqual(results, dataset)

    def test_normalizer_dataset_inplace(self):
        normalizer = BaseNormalizer(lunormalize, attr_index=1)
        dataset = [("a1", "bépoèàÀêùï"), ("a2", "tàtà")]
        normalizer.normalize_dataset(dataset, inplace=True)
        self.assertEqual([["a1", "bepoeaaeui"], ["a2", "tata"]], dataset)

    def test_unormalize(self):
        normalizer = UnicodeNormalizer()
        self.assertEqual(normalizer.normalize("bépoèàÀêùï"), "bepoeaaeui")

    def test_unormalize_record(self):
        normalizer = UnicodeNormalizer(attr_index=1)
        record = ("a1", "bépoèàÀêùï")
        self.assertEqual(["a1", "bepoeaaeui"], normalizer.normalize(record))

    def test_simplify(self):
        normalizer = SimplifyNormalizer(lemmas=FRENCH_LEMMAS)
        self.assertEqual(
            normalizer.normalize("J'aime les frites, les pommes et les scoubidous !"),
            "aimer frites pomme scoubidou",
        )

    def test_simplify_record(self):
        normalizer = SimplifyNormalizer(attr_index=1, lemmas=FRENCH_LEMMAS)
        self.assertEqual(
            ["a1", "aimer frites pomme scoubidou"],
            normalizer.normalize(["a1", "J'aime les frites, les pommes et les scoubidous !"]),
        )

    def test_tokenize(self):
        normalizer = TokenizerNormalizer()
        self.assertEqual(
            [
                "J'",
                "aime",
                "les",
                "frites",
                "!",
            ],
            normalizer.normalize("J'aime les frites !"),
        )

    def test_tokenize_record(self):
        normalizer = TokenizerNormalizer(attr_index=1)
        self.assertEqual(
            [
                "a1",
                [
                    "J'",
                    "aime",
                    "les",
                    "frites",
                    "!",
                ],
            ],
            normalizer.normalize(["a1", "J'aime les frites !"]),
        )

    def test_lemmatizer(self):
        normalizer = LemmatizerNormalizer(FRENCH_LEMMAS)
        self.assertEqual(normalizer.normalize("sacré rubert"), "sacré rubert")
        self.assertEqual(normalizer.normalize("J'aime les frites !"), "je aimer le frite")
        self.assertEqual(normalizer.normalize(", J'aime les frites"), "je aimer le frite")

    def test_lemmatizer_record(self):
        normalizer = LemmatizerNormalizer(FRENCH_LEMMAS, attr_index=1)
        self.assertEqual(["a1", "sacré rubert"], normalizer.normalize(["a1", "sacré rubert"]))
        self.assertEqual(
            ["a1", "je aimer le frite"], normalizer.normalize(["a1", "J'aime les frites !"])
        )
        self.assertEqual(
            ["a1", "je aimer le frite"], normalizer.normalize(["a1", ", J'aime les frites"])
        )

    def test_round(self):
        normalizer = RoundNormalizer()
        self.assertEqual(normalizer.normalize(3.14159), "3")
        normalizer = RoundNormalizer(ndigits=2)
        self.assertEqual(normalizer.normalize(3.14159), "3.14")
        normalizer = RoundNormalizer(ndigits=3)
        self.assertEqual(normalizer.normalize(3.14159), "3.142")

    def test_round_record(self):
        normalizer = RoundNormalizer(attr_index=1)
        self.assertEqual(["a1", "3"], normalizer.normalize(["a1", 3.14159]))
        normalizer = RoundNormalizer(attr_index=1, ndigits=2)
        self.assertEqual(["a1", "3.14"], normalizer.normalize(["a1", 3.14159]))
        normalizer = RoundNormalizer(attr_index=1, ndigits=3)
        self.assertEqual(["a1", "3.142"], normalizer.normalize(["a1", 3.14159]))

    def test_regexp(self):
        normalizer = RegexpNormalizer(r"http://perdu.com/(?P<id>\d+).*", "%(id)s")
        self.assertEqual(normalizer.normalize("http://perdu.com/42/supertop/cool"), "42")

    def test_regexp_record(self):
        normalizer = RegexpNormalizer(r"http://perdu.com/(?P<id>\d+).*", "%(id)s", attr_index=1)
        self.assertEqual(
            ["a1", "42"], normalizer.normalize(["a1", "http://perdu.com/42/supertop/cool"])
        )

    def test_join(self):
        normalizer = JoinNormalizer((1, 2))
        self.assertEqual(normalizer.normalize((1, "ab", "cd", "e", 5)), [1, "e", 5, "ab, cd"])


class NormalizerPipelineTestCase(unittest.TestCase):
    def test_normalizer(self):
        regexp = r'(?P<id>\d+);{["]?(?P<firstname>.+[^"])["]?};{(?P<surname>.*)};{};{};(?P<date>.*)'
        n1 = RegexpNormalizer(regexp, "%(id)s\t%(firstname)s\t%(surname)s\t%(date)s")
        n2 = BaseNormalizer(callback=lambda x: x.split("\t"))
        n3 = UnicodeNormalizer(attr_index=(1, 2, 3))
        pipeline = NormalizerPipeline((n1, n2, n3))
        r1 = '1111;{"Toto tàtà"};{Titi};{};{};'
        self.assertEqual(["1111", "toto tata", "titi", ""], pipeline.normalize(r1))


if __name__ == "__main__":
    unittest.main()



#
# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest
import random
from dateutil import parser as dateparser

from nazca.utils.distances import (
    levenshtein,
    soundex,
    soundexcode,
    difflib_match,
    jaccard,
    euclidean,
    geographical,
    ExactMatchProcessing,
    GeographicalProcessing,
    LevenshteinProcessing,
    SoundexProcessing,
    JaccardProcessing,
    DifflibProcessing,
    DATEUTIL_ENABLED,
)

# Make sure tests are repeatable
random.seed(6)


class DistancesTest(unittest.TestCase):
    def test_difflib_match(self):
        self.assertAlmostEqual(round(difflib_match("Victor Hugo", "Victor Hugo"), 2), 0.0)
        self.assertAlmostEqual(round(difflib_match("Victor Hugo", "Victor Wugo"), 2), 0.09)
        self.assertAlmostEqual(round(difflib_match("niche", "chiens"), 2), 0.45)
        self.assertAlmostEqual(round(difflib_match("bonjour", "bonjour !"), 2), 0.12)

    def test_levenshtein(self):
        self.assertEqual(levenshtein("niche", "chiens"), 5)
        self.assertEqual(levenshtein("bonjour", "bonjour !"), 1)
        self.assertEqual(levenshtein("bon", "bonjour"), 4)
        self.assertEqual(levenshtein("Victor Hugo", "Hugo Victor"), 0)

        # Test symetry
        self.assertEqual(
            levenshtein("Victor Hugo", "Vitor Wugo"), levenshtein("Vitor Wugo", "Victor Hugo")
        )

    def test_soundex(self):
        # Test extracted from Wikipedia en :
        # Using this algorithm :
        # both "Robert" and "Rupert" return the same string "R163"
        # while "Rubin" yields "R150".
        #
        # "Ashcraft" and "Ashcroft" both yield "A261" and not "A226"
        # (the chars 's' and 'c' in the name would receive a single number
        # of 2 and not 22 since an 'h' lies in between them).
        #
        # "Tymczak" yields "T522" not "T520"
        # (the chars 'z' and 'k' in the name are coded as 2 twice since a vowel
        # lies in between them).
        #
        # "Pfister" yields "P236" not "P123" (the first two letters have the same
        # number and are coded once as 'P').

        self.assertEqual(soundexcode("Robert", "english"), "R163")
        self.assertEqual(soundexcode("Rubert", "english"), "R163")
        self.assertEqual(soundexcode("Rubin", "english"), "R150")
        self.assertEqual(soundexcode("Ashcraft", "english"), "A261")
        self.assertEqual(soundexcode("Tymczak", "english"), "T522")
        self.assertEqual(soundexcode("Pfister", "english"), "P236")

        self.assertEqual(soundex("Rubert", "Robert", "english"), 0)
        self.assertEqual(soundex("Rubin", "Robert", "english"), 1)

    def test_jaccard(self):
        # The jaccard indice between two words is the ratio of the number of
        # identical letters and the total number of letters
        # Each letter is counted once only
        # The distance is 1 - jaccard_indice

        self.assertEqual(jaccard("bonjour", "bonjour"), 0.0)
        self.assertAlmostEqual(jaccard("boujour", "bonjour"), 1, 2)
        self.assertAlmostEqual(jaccard("sacré rubert", "sacré hubert"), 0.667, 2)

        # Test symetry
        self.assertEqual(jaccard("orange", "morange"), jaccard("morange", "orange"))

    def test_temporal(self):
        # Test the distance between two dates. The distance can be given in
        # ``days``, ``months`` or ``years``
        try:
            from nazca.distances import temporal
        except ImportError:
            return
        self.assertEqual(temporal("14 aout 1991", "14/08/1991"), 0)
        self.assertEqual(temporal("14 aout 1991", "08/14/1991"), 0)
        self.assertEqual(temporal("14 aout 1991", "08/15/1992"), 367)
        # Test a case of ambiguity
        self.assertEqual(temporal("1er mai 2012", "01/05/2012"), 0)
        self.assertEqual(temporal("1er mai 2012", "05/01/2012", dayfirst=False), 0)
        # Test the different granularities available
        self.assertAlmostEqual(temporal("14 aout 1991", "08/15/1992", "years"), 1.0, 1)
        self.assertAlmostEqual(temporal("1991", "1992", "years"), 1.0, 1)
        self.assertAlmostEqual(temporal("13 mars", "13 mai", "months"), 2.0, 1)
        self.assertAlmostEqual(
            temporal("13 march", "13 may", "months", parserinfo=dateparser.parserinfo), 2.0, 1
        )

        # Test fuzzyness
        self.assertEqual(
            temporal("Jean est né le 1er octobre 1958", "Le 01-10-1958, Jean est né"), 0
        )

        # Test symetry
        self.assertEqual(temporal("14-08-1991", "15/08/1992"), temporal("15/08/1992", "14/08/1991"))

    def test_euclidean(self):
        self.assertEqual(euclidean(10, 11), 1)
        self.assertEqual(euclidean(-10, 11), 21)
        self.assertEqual(euclidean("-10", "11"), 21)

        # Test symetry
        self.assertEqual(euclidean(10, 11), euclidean(11, 10))

    def test_geographical(self):
        paris = (48.856578, 2.351828)
        london = (51.504872, -0.07857)
        dist_parislondon = geographical(paris, london, in_radians=False)
        self.assertAlmostEqual(dist_parislondon, 341564, 0)


class GeographicalTestCase(unittest.TestCase):
    def test_geographical(self):
        # Use the whole record
        processing = GeographicalProcessing(units="km")
        _input = ((48.856578, 2.351828), (51.504872, -0.07857))
        pdist = processing.pdist(_input)
        self.assertEqual([341.56415945105], pdist)

    def test_geographical_2(self):
        # Use a single column of the record (tuple version)
        processing = GeographicalProcessing(ref_attr_index=1, target_attr_index=1, units="km")
        _input = (("paris", (48.856578, 2.351828)), ("london", (51.504872, -0.07857)))
        pdist = processing.pdist(_input)
        self.assertEqual([341.56415945105], pdist)

    def test_geographical_3(self):
        # Use two columns of the record
        processing = GeographicalProcessing(
            ref_attr_index=(1, 2), target_attr_index=(1, 2), units="km"
        )
        _input = (("paris", 48.856578, 2.351828), ("london", 51.504872, -0.07857))
        pdist = processing.pdist(_input)
        self.assertEqual([341.56415945105], pdist)


class ExactMatchTestCase(unittest.TestCase):
    def test_pdist(self):
        processing = ExactMatchProcessing()
        _input = ["Victor Hugo", "Victo Hugo", "Victor Hugo"]
        pdist = processing.pdist(_input)
        self.assertEqual([1, 0.0, 1], pdist)

    def test_index(self):
        processing = ExactMatchProcessing()
        ref_record = ["Victor Hugo", 0]
        target_record = ["Victor Hugo", 1]
        d = processing.distance(ref_record, target_record)
        self.assertEqual(d, 1)

    def test_index_2(self):
        processing = ExactMatchProcessing(ref_attr_index=0, target_attr_index=0)
        ref_record = ["Victor Hugo", 0]
        target_record = ["Victor Hugo", 1]
        d = processing.distance(ref_record, target_record)
        self.assertEqual(d, 0)


class LevenshteinTestCase(unittest.TestCase):
    def setUp(self):
        self.input1 = ["Victor Hugo", "Albert Camus", "Jean Valjean"]
        self.input2 = ["Victor Wugo", "Albert Camus", "Albert Camu"]
        self.distance = levenshtein
        processing = LevenshteinProcessing()
        self.matrix = processing.cdist(self.input1, self.input2)

    def test_matrixconstruction(self):
        d = self.distance
        i1, i2 = self.input1, self.input2
        m = self.matrix

        for i in range(len(i1)):
            for j in range(len(i2)):
                self.assertAlmostEqual(m[i, j], d(i1[i], i2[j]), 4)

    def test_operation(self):
        m = self.matrix
        self.assertTrue((3 * m == m * 3).all())
        self.assertTrue(((m - 0.5 * m) == (0.5 * m)).all())
        self.assertTrue(((m + 10 * m - m * 3) == (8 * m)).all())

    def test_pdist(self):
        _input = ["Victor Wugo", "Albert Camus", "Albert Camu"]
        processing = LevenshteinProcessing()
        pdist = processing.pdist(_input)
        self.assertEqual([6, 6, 1], pdist)


class SoundexTestCase(unittest.TestCase):
    def test_pdist(self):
        processing = SoundexProcessing()
        _input = ["Robert Ugo", "Rubert Ugo", "Rubert Pugo"]
        pdist = processing.pdist(_input)
        self.assertEqual([0, 1, 1], pdist)


class JaccardTestCase(unittest.TestCase):
    def test_pdist(self):
        processing = JaccardProcessing()
        _input = ["Robert Ugo", "Rubert Ugo", "Rubert Pugo"]
        pdist = processing.pdist(_input)
        results = [0.666, 1, 0.666]
        for ind, value in enumerate(pdist):
            self.assertAlmostEqual(results[ind], value, 2)


class DifflibTestCase(unittest.TestCase):
    def test_pdist(self):
        processing = DifflibProcessing()
        _input = ["Robert Ugo", "Rubert Ugo", "Rubert Pugo"]
        pdist = processing.pdist(_input)
        results = [0.099, 0.238, 0.14]
        for ind, value in enumerate(pdist):
            self.assertAlmostEqual(results[ind], value, 2)


@unittest.skipUnless(DATEUTIL_ENABLED, "python-dateutil is not installed")
class TemporalTestCase(unittest.TestCase):
    def test_pdist(self):
        from nazca.utils.distances import TemporalProcessing

        processing = TemporalProcessing()
        _input = ["14 aout 1991", "08/14/1991", "08/15/1992"]
        pdist = processing.pdist(_input)
        self.assertEqual([0.0, 367, 367], pdist)


if __name__ == "__main__":
    unittest.main()



""" Process/Core functions for Named Entities Recognition.
"""

from nazca.utils.tokenizer import RichStringTokenizer


###############################################################################
# NER PROCESS #################################################################
###############################################################################
class NerProcess:
    """High-level process for Named Entities Recognition"""

    def __init__(self, ner_sources, preprocessors=None, filters=None, unique=False):
        """Initialise the class.

        :tokenizer: an instance of tokenizer
        """
        self.ner_sources = list(ner_sources)
        self.preprocessors = preprocessors or []
        self.filters = filters or []
        self.unique = unique

    def add_ner_source(self, process):
        """Add a ner process"""
        self.ner_sources.append(process)

    def add_preprocessors(self, preprocessor):
        """Add a preprocessor"""
        self.preprocessors.append(preprocessor)

    def add_filters(self, filter):
        """Add a filter"""
        self.filters.append(filter)

    def process_text(self, text):
        """High level function for analyzing a text"""
        tokenizer = RichStringTokenizer(text)
        return self.recognize_tokens(tokenizer)

    def recognize_tokens(self, tokens):
        """Recognize Named Entities from a tokenizer or
        an iterator yielding tokens.
        """
        last_stop = 0
        named_entities = []
        for token in tokens:
            if token.start < last_stop:
                continue  # this token overlaps with a previous match
            # Applies preprocessors
            # XXX Preprocessors may be sources dependant
            for preprocessor in self.preprocessors:
                token = preprocessor(token)
                if not token:
                    break
            if not token:
                continue
            recognized = False
            for process in self.ner_sources:
                for uri in process.recognize_token(token):
                    named_entities.append((uri, process.name, token))
                    recognized = True
                    last_stop = token.end
                    if self.unique:
                        break
                if recognized and self.unique:
                    break
        # XXX Postprocess/filters may be sources dependant
        return self.postprocess(named_entities)

    def postprocess(self, named_entities):
        """Postprocess the results by applying filters"""
        for filter in self.filters:
            named_entities = filter(named_entities)
        return named_entities



# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import codecs
import json
import csv
from os.path import exists as fileexists
from os import path as osp

from lxml import etree

try:
    from SPARQLWrapper import SPARQLWrapper, JSON

    SPARQL_ENABLED = True
except ImportError:
    SPARQL_ENABLED = False


###############################################################################
# UTILITY FUNCTIONS ###########################################################
###############################################################################
def autocast(data, encoding=None):
    """Try to convert data into a specific type
    in (int, float, str)
    """
    try:
        return int(data)
    except ValueError:
        try:
            return float(data.replace(",", "."))
        except ValueError:
            data = data.strip()
            if encoding and isinstance(data, bytes):
                return data.decode(encoding)
            return data


###############################################################################
# SPARQL FUNCTIONS ############################################################
###############################################################################
def _sparqlexecute(endpoint, query, raise_on_error=False, agent=None):
    """Execute a sparql query and return the raw results"""
    if not SPARQL_ENABLED:
        raise ImportError(
            "You have to install SPARQLWrapper and JSON modules to used this function"
        )
    if agent:
        sparql = SPARQLWrapper(endpoint, agent=agent)
    else:
        sparql = SPARQLWrapper(endpoint)
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    try:
        try:
            rawresults = sparql.query().convert()
            return rawresults
        except ValueError:
            # Bad json
            rawresults = sparql.query()
            return json.loads(codecs.escape_decode(rawresults.response.read())[0])
    except Exception as err:
        if raise_on_error:
            raise RuntimeError("Error in sparql query", err)
        else:
            return []


def sparqlquery(
    endpoint, query, indexes=None, autocast_data=True, raise_on_error=False, agent=None
):
    """Run the sparql query on the given endpoint, and wrap the items in the
    indexes form. If indexes is empty, keep raw output"""
    results = []
    rawresults = _sparqlexecute(endpoint, query, raise_on_error, agent)
    if not rawresults:
        return results
    labels = rawresults["head"]["vars"]
    indexes = indexes or []
    if autocast_data:
        transform = autocast
    else:
        transform = lambda x: x

    for raw in rawresults["results"]["bindings"]:
        data = []
        if not indexes:
            data = [(transform(raw[label]["value"]) if label in raw else None) for label in labels]
        else:
            for il, ind in enumerate(indexes):
                if isinstance(ind, tuple):
                    data.append(tuple([transform(raw[labels[i]]["value"]) for i in ind]))
                else:
                    data.append(transform(raw[labels[il]]["value"]))
        results.append(data)
    return results


def sparqljson(endpoint, query, lang_order=("fr", "en"), raise_on_error=False, agent=None):
    """Execute and format the results of a sparql query.
    Sort the litterals using lang_order.
    """
    data = {}
    rawresults = _sparqlexecute(endpoint, query, raise_on_error, agent)
    if not rawresults:
        return data
    results = rawresults["results"]["bindings"]
    data_lang = {}
    for row in results:
        for k, v in row.items():
            if v["type"] == "uri":
                # Uri, keep it in a set
                data.setdefault(k, set()).add(v["value"])
            elif v["type"] == "typed-literal":
                # E.g. latitude, longitude, geometry - Keep on value
                data[k] = v["value"]
            else:
                # Literal - Use lang
                data_lang.setdefault(k, []).append((v["value"], v.get("xml:lang")))

    def keyfunc(x):
        return lang_order.index(x[1]) if x[1] in lang_order else len(lang_order)

    data.update({k: sorted(v, key=keyfunc)[0][0] for k, v in data_lang.items()})
    return data


###############################################################################
# FILE FUNCTIONS ##############################################################
###############################################################################
def parsefile(
    filename,
    indexes=None,
    nbmax=None,
    delimiter="\t",
    encoding="utf-8",
    field_size_limit=None,
    autocast_data=True,
    formatopt=None,
):
    """Parse the file (read ``nbmax`` line at maximum if given). Each
    line is splitted according ``delimiter`` and only ``indexes`` are kept

    eg : The file is :
            1, house, 12, 19, apple
            2, horse, 21.9, 19, stramberry
            3, flower, 23, 2.17, cherry

        >>> data = parsefile('myfile', [0, (2, 3), 4, 1], delimiter=',')
        data = [[1, (12, 19), u'apple', u'house'],
                [2, (21.9, 19), u'stramberry', u'horse'],
                [3, (23, 2.17), u'cherry', u'flower']]

        By default, all cells are "autocast" (thanks to the
        ``autocast()`` function), but you can overpass it thanks to the
        ``formatopt`` dictionnary. Each key is the index to work on, and the
        value is the function to call. See the following example:

        >>> data = parsefile('myfile', [0, (2, 3), 4, 1], delimiter=',',
        >>>                  formatopt={2:lambda x:x.decode('utf-8')})
        data = [[1, (u'12', 19), u'apple', u'house'],
                [2, (u'21.9', 19), u'stramberry', u'horse'],
                [3, (u'23', 2.17), u'cherry', u'flower']]

    """

    def formatedoutput(filename):
        if field_size_limit:
            csv.field_size_limit(field_size_limit)

        if isinstance(filename, str):
            csvfile = open(filename)
        else:
            csvfile = filename
        reader = csv.reader(csvfile, delimiter=delimiter)
        for row in reader:
            yield [cell.strip() for cell in row]
        csvfile.close()

    # Autocast if asked
    if autocast_data:

        def deffunc(x):
            return autocast(x, encoding)

    else:

        def deffunc(x):
            return x

    result = []
    indexes = indexes or []
    formatopt = formatopt or {}
    for ind, row in enumerate(formatedoutput(filename)):
        row = [formatopt.get(i, deffunc)(cell) for i, cell in enumerate(row)]
        data = []
        if nbmax and ind > nbmax:
            break
        if not indexes:
            data = row
        else:
            for ind in indexes:
                if isinstance(ind, tuple):
                    data.append(tuple([row[i] for i in ind]))
                    if "" in data[-1]:
                        data[-1] = None
                elif row[ind]:
                    data.append(row[ind])
                else:
                    data.append(None)

        result.append(data)
    return result


def write_results(matched, alignset, targetset, resultfile):
    """Given a matched dictionnay, an alignset and a targetset to the
    resultfile
    """
    openmode = "a" if fileexists(resultfile) else "w"
    with open(resultfile, openmode) as fobj:
        if openmode == "w":
            fobj.write("aligned;targetted;distance\n")
        for aligned in matched:
            for target, dist in matched[aligned]:
                alignid = alignset[aligned][0]
                targetid = targetset[target][0]
                fobj.write(
                    "%s;%s;%s\n"
                    % (
                        alignid,
                        targetid,
                        dist,
                    )
                )


def split_file(filename, outputdir, nblines=60000):
    """Split `filename` into smaller files of ``nblines`` lines. Files are
    written into `outputdir`.

    Return the list of files
    """
    NEW = object()

    def readlines(fobj, nblines):
        """yield all lines of the file, and
        at split-file boundaries, yield a NEW marker
        """
        for index, line in enumerate(fobj):
            if index and index % nblines == 0:
                yield NEW
            yield line

    count = 0
    with open(filename, "rb") as fobj:
        outfile = open(osp.join(outputdir, "%s" % count), "wb")
        for line in readlines(fobj, nblines):
            if line is NEW:
                outfile.close()
                count += 1
                outfile = open(osp.join(outputdir, "%s" % count), "wb")
                continue
            outfile.write(line)
        outfile.close()
        count += 1
    return list(map(str, range(count)))


###############################################################################
# OUTPUT UTILITIES ############################################################
###############################################################################
class AbstractPrettyPrint:
    """Pretty print the output of a named entities process"""

    def pprint_text(self, text, named_entities, **kwargs):
        newtext = ""
        indice = 0
        tindices = {t.start: (uri, t) for uri, p, t in named_entities}
        while indice < len(text):
            if indice in tindices:
                uri, t = tindices[indice]
                words = text[t.start : t.end]
                fragment = self.pprint_entity(uri, words, **kwargs)
                if not self.is_valid(newtext + fragment + text[t.end :]):
                    fragment = words
                newtext += fragment
                indice = t.end
            else:
                newtext += text[indice]
                indice += 1
        return newtext

    def pprint_entity(self, uri, word, **kwargs):
        """Pretty print an entity"""
        raise NotImplementedError

    def is_valid(self, newtext):
        """Override to check the validity of the prettified content at each
        enrichement step"""
        return True


class HTMLPrettyPrint(AbstractPrettyPrint):
    """Pretty print the output of a named entities process, in HTML"""

    def pprint_entity(self, uri, word, **kwargs):
        """Pretty print an entity"""
        klass = ' class="%s"' % kwargs["html_class"] if "html_class" in kwargs else ""
        return '<a href="{}"{}>{}</a>'.format(uri, klass, word)


class ValidXHTMLPrettyPrint(HTMLPrettyPrint):
    """Pretty print the output of a named entities process,
    in valid XHTML.
    """

    XHTML_DOC_TEMPLATE = b"""\
<?xml version="1.0" encoding="UTF-8" ?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<meta http-equiv="content-type" content="text/html; charset=UTF-8"/>
<title>ner</title>
</head>
<body><div>%s</div></body>
</html>"""  # noqa

    def is_valid(self, html):
        if isinstance(html, str):
            html = html.encode("utf-8")
        try:
            etree.fromstring(
                self.XHTML_DOC_TEMPLATE % html,
                parser=etree.XMLParser(dtd_validation=True, no_network=False),
            )
        except etree.XMLSyntaxError:
            return False
        return True



#!/usr/bin/python

from os import path

import six.moves.urllib.request, six.moves.urllib.parse, six.moves.urllib.error

import nazca.distances as ald
import nazca.normalize as aln
from nazca.aligner import align, subalign, findneighbours, alignall
from nazca.dataio import parsefile, sparqlquery, write_results

DEMODIR = path.dirname(__file__)


def dpath(filename):
    return path.join(DEMODIR, "demo", filename)


def remove_after(string, sub):
    try:
        return string[: string.lower().index(sub)].strip()
    except ValueError:
        return string


def parserql(host, rql):
    filehandle = six.moves.urllib.request.urlopen(f"{host}view?rql={rql}&vid=csvexport")
    filehandle.readline()
    rset = [[e.decode("utf-8") for e in line.strip().split(";")] for line in filehandle]
    return rset


def demo_0():
    # prixgoncourt is the list of Goncourt Prize, extracted
    # from wikipedia

    # We try to align Goncourt winers onto dbpedia results

    query = """
       SELECT ?writer, ?name WHERE {
          ?writer  <http://purl.org/dc/terms/subject> <http://dbpedia.org/resource/Category:French_novelists>.
          ?writer rdfs:label ?name.
          FILTER(lang(?name) = 'fr')
       }
    """

    print("Sending query to dbpedia")
    targetset = sparqlquery("http://dbpedia.org/sparql", query)
    print("Reading the prixgoncourt file")
    alignset = parsefile(dpath("prixgoncourt"), indexes=[1, 1])

    tr_name = {
        "normalization": [lambda x: remove_after(x, "("), aln.simplify],
        "metric": ald.levenshtein,
    }

    processings = {1: tr_name}

    print("Alignment started")
    align(alignset, targetset, 0.4, processings, dpath("demo0_results"))

    print("Done, see the resuls in %s" % dpath("demo0_results"))


def demo_1():
    # FR.txt is an extract of geonames, where locations have been sorted by name
    # frenchbnf is an extract of french BNF's locations, sorted by name too

    # For each line (ie location) we keep the identifier, the name and the
    # position (longitude, latitude)
    # ``nbmax`` is the number of locations to load

    print("Parsing the input files")
    targetset = parsefile(dpath("FR.txt"), indexes=[0, 1, (4, 5)], nbmax=2000)
    alignset = parsefile(dpath("frenchbnf"), indexes=[0, 2, (14, 12)], nbmax=1000)

    # Let's define the processings to apply on the location's name
    tr_name = {
        "normalization": [aln.simplify],  # Simply all the names (remove
        #   punctuation, lower case, etc)
        "metric": ald.levenshtein,  # Use the levenshtein distance
        "weighting": 1,  # Use 1 a name-distance matrix
        #   weighting coefficient
    }
    tr_geo = {
        "normalization": [],  # No normalization needed
        "metric": ald.geographical,  # Use the geographical distance
        "metric_params": {"units": "km"},  # Arguments given the
        #   distance function. Here,
        #   the unit to use
        "weighting": 1,
    }

    processings = {1: tr_name, 2: tr_geo}

    print("Alignment started")
    align(
        alignset,  # The dataset to align
        targetset,  # The target dataset
        0.4,  # The maximal distance
        #   threshold
        processings,  # The list of processings to
        #   apply.
        dpath("demo1_results"),
    )
    # Filename of the output
    #   result file
    # the ``align()`` function return two items
    # 0. the computed distance matrix
    # 1. a boolean, True if at least one alignment has been done, False
    #    otherwise
    print("Done, see the results in %s" % dpath("demo1_results"))


def demo_2():
    targetset = parsefile(
        dpath("FR.txt"), indexes=[0, 1, (4, 5)], formatopt={1: lambda x: x.decode("utf-8")}
    )
    alignset = parsefile(
        dpath("frenchbnf"),
        indexes=[0, 2, (14, 12)],
        formatopt={2: lambda x: x.decode("utf-8")},
        nbmax=30000,
    )

    print("Finding neighbours")
    neighbours = findneighbours(alignset, targetset, indexes=(2, 2), mode="minibatch")

    # Let's define the processings to apply on the location's name
    tr_name = {
        "normalization": [aln.simplify],  # Simply all the names (remove
        #   punctuation, lower case, etc)
        "metric": ald.levenshtein,  # Use the levenshtein distance
        "weighting": 1,  # Use 1 a name-distance matrix
        #   weighting coefficient
    }

    processings = {1: tr_name}
    print("Start computation")
    for ind, (alignid, targetid) in enumerate(neighbours):
        print("%3d" % ind, len(alignid), "x", len(targetid))
        _, matched = subalign(
            alignset,  # The dataset to align
            targetset,  # The target dataset
            alignid,
            targetid,
            0.3,
            processings,
        )
        write_results(matched, alignset, targetset, dpath("demo2_results"))
    print("Done, see the results in %s" % dpath("demo2_results"))


def demo_3():
    print("Parsing files")
    alignset = parserql(
        host="http://demo.cubicweb.org/elections/",
        rql="Any E, N WHERE X is Commune, X eid E, X label N",
    )
    targetset = parsefile(dpath("FR.txt"), indexes=[0, 1])
    print(f"{len(alignset)}×{len(targetset)}")

    tr_name = {"normalization": [aln.simplify], "metric": "levenshtein"}

    print("Alignment started")
    results = alignall(
        alignset,
        targetset,
        0.75,
        processings={1: tr_name},
        indexes=(1, 1),
        mode="minhashing",
        kwordsgram=1,
        siglen=200,
        uniq=True,
    )
    dicresults = {a: b for (a, b) in results}

    print("Done, writing output")

    with open(dpath("demo3_results"), "w") as fout:
        for line in alignset:
            sent = (
                "http://demo.cubicweb.org/elections/commune/%s;"
                "http://www.geonames.org/%s\n" % (line[0], dicresults.get(line[0], "not_found"))
            )
            fout.write(sent.encode("utf-8"))

    print("See the results in %s" % dpath("demo3_results"))


if __name__ == "__main__":
    import sys
    from time import time

    runall = len(sys.argv) == 1

    t = time()
    if runall or "0" in sys.argv:
        print("Running demo_0")
        demo_0()

    if runall or "1" in sys.argv:
        print("Running demo_1")
        demo_1()

    if runall or "2" in sys.argv:
        print("Running demo_2")
        ## Same as demo_1, but in a more efficient way, using a method to find
        ## neighbours
        demo_2()

    if runall or "3" in sys.argv:
        print("Running demo_3")
        demo_3()

    print("Demo terminated")
    print("Took %d min" % ((time() - t) / 60.0))
